# vim: expandtab:ts=4:sw=4
import argparse
import tensorflow as tf
import tensorflow.contrib.slim as slim


def _batch_norm_fn(x, scope=None):
    if scope is None:
        scope = tf.get_variable_scope().name + "/bn"
    return slim.batch_norm(x, scope=scope)


def create_link(
        incoming, network_builder, scope, nonlinearity=tf.nn.elu,
        weights_initializer=tf.truncated_normal_initializer(stddev=1e-3),
        regularizer=None, is_first=False, summarize_activations=True):
    if is_first:
        network = incoming
    else:
        network = _batch_norm_fn(incoming, scope=scope + "/bn")
        network = nonlinearity(network)
        if summarize_activations:
            tf.summary.histogram(scope+"/activations", network)

    pre_block_network = network
    post_block_network = network_builder(pre_block_network, scope)

    incoming_dim = pre_block_network.get_shape().as_list()[-1]
    outgoing_dim = post_block_network.get_shape().as_list()[-1]
    if incoming_dim != outgoing_dim:
        assert outgoing_dim == 2 * incoming_dim, \
            "%d != %d" % (outgoing_dim, 2 * incoming)
        projection = slim.conv2d(
            incoming, outgoing_dim, 1, 2, padding="SAME", activation_fn=None,
            scope=scope+"/projection", weights_initializer=weights_initializer,
            biases_initializer=None, weights_regularizer=regularizer)
        network = projection + post_block_network
    else:
        network = incoming + post_block_network
    return network


def create_inner_block(
        incoming, scope, nonlinearity=tf.nn.elu,
        weights_initializer=tf.truncated_normal_initializer(1e-3),
        bias_initializer=tf.zeros_initializer(), regularizer=None,
        increase_dim=False, summarize_activations=True):
    n = incoming.get_shape().as_list()[-1]
    stride = 1
    if increase_dim:
        n *= 2
        stride = 2

    incoming = slim.conv2d(
        incoming, n, [3, 3], stride, activation_fn=nonlinearity, padding="SAME",
        normalizer_fn=_batch_norm_fn, weights_initializer=weights_initializer,
        biases_initializer=bias_initializer, weights_regularizer=regularizer,
        scope=scope + "/1")
    if summarize_activations:
        tf.summary.histogram(incoming.name + "/activations", incoming)

    incoming = slim.dropout(incoming, keep_prob=0.6)

    incoming = slim.conv2d(
        incoming, n, [3, 3], 1, activation_fn=None, padding="SAME",
        normalizer_fn=None, weights_initializer=weights_initializer,
        biases_initializer=bias_initializer, weights_regularizer=regularizer,
        scope=scope + "/2")
    return incoming


def residual_block(incoming, scope, nonlinearity=tf.nn.elu,
                   weights_initializer=tf.truncated_normal_initializer(1e3),
                   bias_initializer=tf.zeros_initializer(), regularizer=None,
                   increase_dim=False, is_first=False,
                   summarize_activations=True):

    def network_builder(x, s):
        return create_inner_block(
            x, s, nonlinearity, weights_initializer, bias_initializer,
            regularizer, increase_dim, summarize_activations)

    return create_link(
        incoming, network_builder, scope, nonlinearity, weights_initializer,
        regularizer, is_first, summarize_activations)


def _create_network(incoming, reuse=None, weight_decay=1e-8):
    nonlinearity = tf.nn.elu
    conv_weight_init = tf.truncated_normal_initializer(stddev=1e-3)
    conv_bias_init = tf.zeros_initializer()
    conv_regularizer = slim.l2_regularizer(weight_decay)
    fc_weight_init = tf.truncated_normal_initializer(stddev=1e-3)
    fc_bias_init = tf.zeros_initializer()
    fc_regularizer = slim.l2_regularizer(weight_decay)

    def batch_norm_fn(x):
        return slim.batch_norm(x, scope=tf.get_variable_scope().name + "/bn")

    network = incoming
    network = slim.conv2d(
        network, 32, [3, 3], stride=1, activation_fn=nonlinearity,
        padding="SAME", normalizer_fn=batch_norm_fn, scope="conv1_1",
        weights_initializer=conv_weight_init, biases_initializer=conv_bias_init,
        weights_regularizer=conv_regularizer)
    network = slim.conv2d(
        network, 32, [3, 3], stride=1, activation_fn=nonlinearity,
        padding="SAME", normalizer_fn=batch_norm_fn, scope="conv1_2",
        weights_initializer=conv_weight_init, biases_initializer=conv_bias_init,
        weights_regularizer=conv_regularizer)

    # NOTE(nwojke): This is missing a padding="SAME" to match the CNN
    # architecture in Table 1 of the paper. Information on how this affects
    # performance on MOT 16 training sequences can be found in
    # issue 10 https://github.com/nwojke/deep_sort/issues/10
    network = slim.max_pool2d(network, [3, 3], [2, 2], scope="pool1")

    network = residual_block(
        network, "conv2_1", nonlinearity, conv_weight_init, conv_bias_init,
        conv_regularizer, increase_dim=False, is_first=True)
    network = residual_block(
        network, "conv2_3", nonlinearity, conv_weight_init, conv_bias_init,
        conv_regularizer, increase_dim=False)

    network = residual_block(
        network, "conv3_1", nonlinearity, conv_weight_init, conv_bias_init,
        conv_regularizer, increase_dim=True)
    network = residual_block(
        network, "conv3_3", nonlinearity, conv_weight_init, conv_bias_init,
        conv_regularizer, increase_dim=False)

    network = residual_block(
        network, "conv4_1", nonlinearity, conv_weight_init, conv_bias_init,
        conv_regularizer, increase_dim=True)
    network = residual_block(
        network, "conv4_3", nonlinearity, conv_weight_init, conv_bias_init,
        conv_regularizer, increase_dim=False)

    feature_dim = network.get_shape().as_list()[-1]
    network = slim.flatten(network)

    network = slim.dropout(network, keep_prob=0.6)
    network = slim.fully_connected(
        network, feature_dim, activation_fn=nonlinearity,
        normalizer_fn=batch_norm_fn, weights_regularizer=fc_regularizer,
        scope="fc1", weights_initializer=fc_weight_init,
        biases_initializer=fc_bias_init)

    features = network

    # Features in rows, normalize axis 1.
    features = slim.batch_norm(features, scope="ball", reuse=reuse)
    feature_norm = tf.sqrt(
        tf.constant(1e-8, tf.float32) +
        tf.reduce_sum(tf.square(features), [1], keepdims=True))
    features = features / feature_norm
    return features, None


def _network_factory(weight_decay=1e-8):

    def factory_fn(image, reuse):
            with slim.arg_scope([slim.batch_norm, slim.dropout],
                                is_training=False):
                with slim.arg_scope([slim.conv2d, slim.fully_connected,
                                     slim.batch_norm, slim.layer_norm],
                                    reuse=reuse):
                    features, logits = _create_network(
                        image, reuse=reuse, weight_decay=weight_decay)
                    return features, logits

    return factory_fn


def _preprocess(image):
    image = image[:, :, ::-1]  # BGR to RGB
    return image


def parse_args():
    """Parse command line arguments.
    """
    parser = argparse.ArgumentParser(description="Freeze old model")
    parser.add_argument(
        "--checkpoint_in",
        default="resources/networks/mars-small128.ckpt-68577",
        help="Path to checkpoint file")
    parser.add_argument(
        "--graphdef_out",
        default="resources/networks/mars-small128.pb")
    return parser.parse_args()


def main():
    args = parse_args()

    with tf.Session(graph=tf.Graph()) as session:
        input_var = tf.placeholder(
            tf.uint8, (None, 128, 64, 3), name="images")
        image_var = tf.map_fn(
            lambda x: _preprocess(x), tf.cast(input_var, tf.float32),
            back_prop=False)

        factory_fn = _network_factory()
        features, _ = factory_fn(image_var, reuse=None)
        features = tf.identity(features, name="features")

        saver = tf.train.Saver(slim.get_variables_to_restore())
        saver.restore(session, args.checkpoint_in)

        output_graph_def = tf.graph_util.convert_variables_to_constants(
            session, tf.get_default_graph().as_graph_def(),
            [features.name.split(":")[0]])
        with tf.gfile.GFile(args.graphdef_out, "wb") as file_handle:
            file_handle.write(output_graph_def.SerializeToString())


if __name__ == "__main__":
    main()



"""
MODIFIED FROM keras-yolo3 PACKAGE, https://github.com/qqwweee/keras-yolo3
Retrain the YOLO model for your own dataset.
"""


import os
import sys


import numpy as np
import keras.backend as K
from keras.layers import Input, Lambda
from keras.models import Model
from keras.optimizers import Adam
from keras.callbacks import (
    TensorBoard,
    ModelCheckpoint,
    ReduceLROnPlateau,
    EarlyStopping,
)
from .yolo3.model import (
    preprocess_true_boxes,
    yolo_body,
    tiny_yolo_body,
    yolo_loss,
)
from .yolo3.utils import get_random_data
from PIL import Image


def get_classes(classes_path):
    """loads the classes"""
    with open(classes_path) as f:
        class_names = f.readlines()
    class_names = [c.strip() for c in class_names]
    return class_names


def get_anchors(anchors_path):
    """loads the anchors from a file"""
    with open(anchors_path) as f:
        anchors = f.readline()
    anchors = [float(x) for x in anchors.split(",")]
    return np.array(anchors).reshape(-1, 2)


def create_model(
    input_shape,
    anchors,
    num_classes,
    load_pretrained=True,
    freeze_body=2,
    weights_path="keras_yolo3/model_data/yolo_weights.h5",
):
    """create the training model"""
    K.clear_session()  # get a new session
    image_input = Input(shape=(None, None, 3))
    h, w = input_shape
    num_anchors = len(anchors)

    y_true = [
        Input(
            shape=(
                h // {0: 32, 1: 16, 2: 8}[l],
                w // {0: 32, 1: 16, 2: 8}[l],
                num_anchors // 3,
                num_classes + 5,
            )
        )
        for l in range(3)
    ]

    model_body = yolo_body(image_input, num_anchors // 3, num_classes)
    print(
        "Create YOLOv3 model with {} anchors and {} classes.".format(
            num_anchors, num_classes
        )
    )

    if load_pretrained:
        model_body.load_weights(weights_path, by_name=True, skip_mismatch=True)
        print("Load weights {}.".format(weights_path))
        if freeze_body in [1, 2]:
            # Freeze darknet53 body or freeze all but 3 output layers.
            num = (185, len(model_body.layers) - 3)[freeze_body - 1]
            for i in range(num):
                model_body.layers[i].trainable = False
            print(
                "Freeze the first {} layers of total {} layers.".format(
                    num, len(model_body.layers)
                )
            )

    model_loss = Lambda(
        yolo_loss,
        output_shape=(1,),
        name="yolo_loss",
        arguments={
            "anchors": anchors,
            "num_classes": num_classes,
            "ignore_thresh": 0.5,
        },
    )([*model_body.output, *y_true])
    model = Model([model_body.input, *y_true], model_loss)

    return model


def create_tiny_model(
    input_shape,
    anchors,
    num_classes,
    load_pretrained=True,
    freeze_body=2,
    weights_path="keras_yolo3/model_data/tiny_yolo_weights.h5",
):
    """create the training model, for Tiny YOLOv3"""
    K.clear_session()  # get a new session
    image_input = Input(shape=(None, None, 3))
    h, w = input_shape
    num_anchors = len(anchors)

    y_true = [
        Input(
            shape=(
                h // {0: 32, 1: 16}[l],
                w // {0: 32, 1: 16}[l],
                num_anchors // 2,
                num_classes + 5,
            )
        )
        for l in range(2)
    ]

    model_body = tiny_yolo_body(image_input, num_anchors // 2, num_classes)
    print(
        "Create Tiny YOLOv3 model with {} anchors and {} classes.".format(
            num_anchors, num_classes
        )
    )

    if load_pretrained:
        model_body.load_weights(weights_path, by_name=True, skip_mismatch=True)
        print("Load weights {}.".format(weights_path))
        if freeze_body in [1, 2]:
            # Freeze the darknet body or freeze all but 2 output layers.
            num = (20, len(model_body.layers) - 2)[freeze_body - 1]
            for i in range(num):
                model_body.layers[i].trainable = False
            print(
                "Freeze the first {} layers of total {} layers.".format(
                    num, len(model_body.layers)
                )
            )

    model_loss = Lambda(
        yolo_loss,
        output_shape=(1,),
        name="yolo_loss",
        arguments={
            "anchors": anchors,
            "num_classes": num_classes,
            "ignore_thresh": 0.7,
        },
    )([*model_body.output, *y_true])
    model = Model([model_body.input, *y_true], model_loss)

    return model


def data_generator(annotation_lines, batch_size, input_shape, anchors, num_classes):
    """data generator for fit_generator"""
    n = len(annotation_lines)
    i = 0
    while True:
        image_data = []
        box_data = []
        for b in range(batch_size):
            if i == 0:
                np.random.shuffle(annotation_lines)
            image, box = get_random_data(annotation_lines[i], input_shape, random=True)
            image_data.append(image)
            box_data.append(box)
            i = (i + 1) % n
        image_data = np.array(image_data)
        box_data = np.array(box_data)
        y_true = preprocess_true_boxes(box_data, input_shape, anchors, num_classes)
        yield [image_data, *y_true], np.zeros(batch_size)


def data_generator_wrapper(
    annotation_lines, batch_size, input_shape, anchors, num_classes
):
    n = len(annotation_lines)
    if n == 0 or batch_size <= 0:
        return None
    return data_generator(
        annotation_lines, batch_size, input_shape, anchors, num_classes
    )


def ChangeToOtherMachine(filelist, repo="TrainYourOwnYOLO", remote_machine=""):
    """
    Takes a list of file_names located in a repo and changes it to the local machines file names. File must be executed from withing the repository

    Example:

    '/home/ubuntu/TrainYourOwnYOLO/Data/Street_View_Images/vulnerable/test.jpg'

    Get's converted to
    
    'C:/Users/Anton/TrainYourOwnYOLO/Data/Street_View_Images/vulnerable/test.jpg'

    """
    filelist = [x.replace("\\", "/") for x in filelist]
    if repo[-1] == "/":
        repo = repo[:-1]
    if remote_machine:
        prefix = remote_machine.replace("\\", "/")
    else:
        prefix = ((os.getcwd().split(repo))[0]).replace("\\", "/")
    new_list = []

    for file in filelist:
        suffix = (file.split(repo))[1]
        if suffix[0] == "/":
            suffix = suffix[1:]
        new_list.append(os.path.join(prefix, repo + "/", suffix).replace("\\", "/"))
    print(
        "8888888888888888888*********************************98888888888888888888888888888888888"
    )
    print(new_list)
    return new_list



""" This Python Script Contains all Predefined ImageNet Models with Their Implementations

which will be directed towards Classification File

Author : Tushar Goel

"""
import tensorflow as tf
import os
import wget
import sys


class Create_Model:
    """
    This Class contains functions of Different models which will return model
    
    Arguments:
        working directory --> Working Directory where Material is present
        image_shape --> Default Shape: (224,224,3)
        train --> False(default), whether to predict or train the model
        
    Methods:
        Different Architecture Model:
            InceptionV3
            Xception
            VGG16
            ResNet50
            VGG19
    """
    
    def __init__(self,working_directory,image_shape=(224,224,3),train=False,input_tensor=None):
        
        self.working_directory = working_directory      # Working Directory
        self.train = train                              # Train or Predicting
        self.image_shape = image_shape                  # Image Shape
        self.input_tensor=input_tensor
            
    def MobileNetV2(self):
        """
        Initiatisation of Mobile Net V2 Model
        
        """
        
        if self.train:
            weights_no_top = os.path.join(self.working_directory,'mobilenet_notop_model.h5')
            
            if not os.path.exists(weights_no_top):
                print('Downloading MobileNetV2_Notop_weights...')
                url = 'https://github.com/JonathanCMitchell/mobilenet_v2_keras/releases/download/v1.1/mobilenet_v2_weights_tf_dim_ordering_tf_kernels_0.5_224_no_top.h5'
                wget.download(url,weights_no_top,bar=bar_progress)
                print('\n---------Weight Downloaded-------------')
            print('\nWeights Loaded\n')
            model = tf.keras.applications.MobileNetV2(include_top=False,input_shape=self.image_shape,weights=weights_no_top)
            
        else:
            weights_top = os.path.join(self.working_directory,'mobilenet_model.h5')
            
            if not os.path.exists(weights_top):
                print('Downloading MobileNetV2_weights...')
                url = 'https://github.com/JonathanCMitchell/mobilenet_v2_keras/releases/download/v1.1/mobilenet_v2_weights_tf_dim_ordering_tf_kernels_0.5_224.h5'
                wget.download(url,weights_top,bar=bar_progress)
                print('\n---------Weight Downloaded-------------')
            print('\nWeights Loaded\n')
            model = tf.keras.applications.MobileNetV2(include_top=True,weights=weights_top,input_tensor=self.input_tensor)
        return model
    
    def InceptionV3(self):
        """
        Initiatisation of InceptionV3 Model
        
        """
        
        if self.train:
            weights_no_top = os.path.join(self.working_directory,'inceptionv3_notop_model.h5')
            
            if not os.path.exists(weights_no_top):
                print('Downloading InceptionV3_Notop_weights...')
                url = 'https://github.com/fchollet/deep-learning-models/releases/download/v0.5/inception_v3_weights_tf_dim_ordering_tf_kernels_notop.h5'
                wget.download(url,weights_no_top,bar=bar_progress)
                print('\n---------Weight Downloaded-------------')
            print('\nWeights Loaded\n')
            model = tf.keras.applications.InceptionV3(include_top=False,input_shape=self.image_shape,weights=weights_no_top)
            
        else:
            weights_top = os.path.join(self.working_directory,'inceptionv3_model.h5')
            
            if not os.path.exists(weights_top):
                print('Downloading InceptionV3_weights...')
                url = 'https://github.com/fchollet/deep-learning-models/releases/download/v0.5/inception_v3_weights_tf_dim_ordering_tf_kernels.h5'
                wget.download(url,weights_top,bar=bar_progress)
                print('\n---------Weight Downloaded-------------')
            print('\nWeights Loaded\n')
            model = tf.keras.applications.InceptionV3(include_top=True,weights=weights_top,input_tensor=self.input_tensor)
        return model
        
    def ResNet50(self):
        """
        Intialisation of Resnet50 Model
        
        """
        if self.train:
            weights_no_top = os.path.join(self.working_directory,'ResNet50_notop_model.h5')
            
            if not os.path.exists(weights_no_top):
                print('Downloading Resnet50_Notop_weights...')
                url = 'https://github.com/fchollet/deep-learning-models/releases/download/v0.2/resnet50_weights_tf_dim_ordering_tf_kernels_notop.h5'
                wget.download(url,weights_no_top,bar=bar_progress)
                print('\n---------Weight Downloaded-------------')
            print('\nWeights Loaded\n')
            model = tf.keras.applications.ResNet50(include_top=False,input_shape=self.image_shape,weights=weights_no_top,)
            
        else:
            weights_top = os.path.join(self.working_directory,'ResNet50_model.h5')
            
            if not os.path.exists(weights_top):
                print('Downloading Resnet50_weights...')
                url = 'https://github.com/fchollet/deep-learning-models/releases/download/v0.2/resnet50_weights_tf_dim_ordering_tf_kernels.h5'
                wget.download(url,weights_top,bar=bar_progress)
                print('\n---------Weight Downloaded-------------')
            print('\nWeights Loaded\n')
            model = tf.keras.applications.ResNet50(include_top=True,weights=weights_top,input_tensor=self.input_tensor)
        return model
    
    def Xception(self):
        """
        Initialisation of Xception Model
        
        """
        if self.train:
            weights_no_top = os.path.join(self.working_directory,'Xception_notop_model.h5')
            
            if not os.path.exists(weights_no_top):
                print('Downloading Xception_Notop_weights...')
                url = 'https://github.com/fchollet/deep-learning-models/releases/download/v0.4/xception_weights_tf_dim_ordering_tf_kernels_notop.h5'
                wget.download(url,weights_no_top,bar=bar_progress)
                print('\n---------Weight Downloaded-------------')
            print('\nWeights Loaded\n')
            model = tf.keras.applications.Xception(include_top=False,input_shape=self.image_shape,weights=weights_no_top)
            
        else:
            weights_top = os.path.join(self.working_directory,'Xception_model.h5')
            
            if not os.path.exists(weights_top):
                print('Downloading Xception_weights...')
                url = 'https://github.com/fchollet/deep-learning-models/releases/download/v0.4/xception_weights_tf_dim_ordering_tf_kernels.h5'
                wget.download(url,weights_top,bar=bar_progress)
                print('\n---------Weight Downloaded-------------')
            print('\nWeights Loaded\n')
            model = tf.keras.applications.Xception(include_top=True,weights=weights_top,input_tensor=self.input_tensor)
        return model
    
    def VGG16(self):
        """
        Initialisation of VGG16 Model
        
        """
        if self.train:
            weights_no_top = os.path.join(self.working_directory,'VGG16_notop_model.h5')
            
            if not os.path.exists(weights_no_top):
                print('Downloading VGG16_Notop_weights...')
                url = 'https://github.com/fchollet/deep-learning-models/releases/download/v0.1/vgg16_weights_tf_dim_ordering_tf_kernels_notop.h5'
                wget.download(url,weights_no_top,bar=bar_progress)
                print('\n---------Weight Downloaded-------------')
            print('\nWeights Loaded\n')
            model = tf.keras.applications.VGG16(include_top=False,input_shape=self.image_shape,weights=weights_no_top)
            
        else:
            weights_top = os.path.join(self.working_directory,'VGG16_model.h5')
            
            if not os.path.exists(weights_top):
                print('Downloading VGG16_weights...')
                url = 'https://github.com/fchollet/deep-learning-models/releases/download/v0.1/vgg16_weights_tf_dim_ordering_tf_kernels.h5'
                wget.download(url,weights_top,bar=bar_progress)
                print('\n---------Weight Downloaded-------------')
            print('\nWeights Loaded\n')
            model = tf.keras.applications.VGG16(include_top=True,weights=weights_top,input_tensor=self.input_tensor)
        return model

    def VGG19(self):
        """
        Initialisation of VGG16 Model
        
        """
        if self.train:
            weights_no_top = os.path.join(self.working_directory,'VGG19_notop_model.h5')
            
            if not os.path.exists(weights_no_top):
                print('Downloading VGG19_Notop_weights...')
                url = 'https://github.com/fchollet/deep-learning-models/releases/download/v0.1/vgg19_weights_tf_dim_ordering_tf_kernels_notop.h5'
                wget.download(url,weights_no_top,bar=bar_progress)
                print('\n---------Weight Downloaded-------------')
            print('\nWeights Loaded\n')
            model = tf.keras.applications.VGG19(include_top=False,input_shape=self.image_shape,weights=weights_no_top)
            
        else:
            weights_top = os.path.join(self.working_directory,'VGG19_model.h5')
            
            if not os.path.exists(weights_top):
                print('Downloading VGG19_weights...')
                url = 'https://github.com/fchollet/deep-learning-models/releases/download/v0.1/vgg19_weights_tf_dim_ordering_tf_kernels.h5'
                wget.download(url,weights_top,bar=bar_progress)
                print('\n---------Weight Downloaded-------------')
            print('\nWeights Loaded\n')
            model = tf.keras.applications.VGG19(include_top=True,weights=weights_top,input_tensor=self.input_tensor)
        return model


def Train_Model(model,num_classes,train_data_object,working_directory,output_directory,optimizer,loss,epochs,metrics,validation_data_object=None,fine_tuning=False,
                layers = 20 , save_model = True,validation_steps=80,rebuild=False,steps_per_epoch = 50,callbacks=False):
    """
    This Function will be used to train the Model and save the model to Output Directory
    
    Arguments:
        Model --> This will be the Initiated Model returned from Create Model Class
        fine_tuning --> whether to unfreeze the layers and tune them
        layers --> no of layers to unfreeze : 20(Default)
        save_model --> True(Default) whether to save Model or not
        train_data_object --> Training Data Generated from Preprocessing the Function
        
    Returns :
        Trained Model
    """
    
    log_directory = os.path.join(working_directory,'\logs')
    model_checkpoint_directory = os.path.join(working_directory,'\model_checkpoint')            
    if not os.path.exists(log_directory):
        os.mkdir(log_directory)
    
    if not os.path.exists(model_checkpoint_directory):
        os.mkdir(model_checkpoint_directory)
    if num_classes == 2:
            target = num_classes - 1
            activation = 'sigmoid'
    else:
            target = num_classes
            activation = 'softmax'
    if rebuild:
        
        model.trainable = False    
        if fine_tuning:
            #Fine tuning for Increasing the accuracy of the model
            
            model.trainable = True #Unfreeze all layer
            layers_length = len(model.layers)
            
            if layers == 'all':
                model.trainable = True
            # Lets Freeze the Bottom Layers:
            else:
                freeze_layer_length = layers_length - layers
                for layers in model.layers[:freeze_layer_length]:
                    layers.trainable = False
                
                for layers in model.layers[freeze_layer_length:]:
                    layers.trainable = True
    
        
        
        layer1 = tf.keras.layers.GlobalAveragePooling2D()
        output = tf.keras.layers.Dense(target,activation=activation)
        print('Build Successfully')
        N_Model = tf.keras.models.Sequential()
        N_Model.add(model)
        N_Model.add(layer1)
        N_Model.add(output)
        print(N_Model.summary())
        New_Model = N_Model
    else:
        New_Model = model
        if New_Model.output_shape[1] != target:
            raise ValueError('Correct the Number of Classes')
    
    if callbacks:
        my_callbacks = callbacks
        
        New_Model.compile(loss = loss,optimizer = optimizer,metrics = metrics)
        if validation_data_object is None:
            
            history = New_Model.fit_generator(train_data_object,steps_per_epoch = steps_per_epoch,epochs= epochs, callbacks = my_callbacks)
            
        else:
            
            history = New_Model.fit_generator(train_data_object,steps_per_epoch = steps_per_epoch,epochs=epochs,validation_data = validation_data_object,
                                callbacks = my_callbacks)
    else:
        New_Model.compile(loss = loss,optimizer = optimizer,metrics = metrics)
        if validation_data_object is None:
            
            history = New_Model.fit_generator(train_data_object,steps_per_epoch = steps_per_epoch,epochs= epochs)
            
        else:
            
            history = New_Model.fit_generator(train_data_object,steps_per_epoch = steps_per_epoch,epochs=epochs,validation_data = validation_data_object,validation_steps=validation_steps)
        
    return history,New_Model

 
def bar_progress(current, total, width=80):
  progress_message = "Downloading: %d%% [%d / %d] bytes" % (current / total * 100, current, total)
  # Don't use print() as it will print in new line every time.
  sys.stdout.write("\r" + progress_message)
  sys.stdout.flush()
        



from PIL import Image
import numpy as np
import cv2
from .yolo import YOLO
import os 
import pyautogui

width,height = pyautogui.size()[0],pyautogui.size()[1]

def Real_Time_Tracking(working_directory,classes = [],model_name=None,score = 0.6,iou = 0.5,gpu_num = 1):
    
    anchors_path = os.path.join(os.path.dirname(__file__),'yolo4_anchors.txt')
    if model_name is None:
        model_path = os.path.join(working_directory,'yolov4.h5')
        classes_path = os.path.join(os.path.dirname(__file__),'coco_classes.txt')
    else:
        model_path = os.path.join(working_directory,model_name)
        classes_path = os.path.join(working_directory,'data_classes.txt')
    
    yolo = YOLO(
        **{
            "model_path": model_path,
            "anchors_path": anchors_path,
            "classes_path": classes_path,
            "score": score,
            "gpu_num": gpu_num,
            "model_image_size": (608,608),
        }
    )
    out = cv2.VideoWriter(os.path.join(working_directory,'output.avi'),cv2.VideoWriter_fourcc('X','V','I','D'), 10, (width,height))
    while True:
        # make a screenshot
        img = pyautogui.screenshot()
        # convert these pixels to a proper numpy array to work with OpenCV
        prediction,img = yolo.detect_image(img,classes = classes,score = score,show_stats=False)
        frame = np.array(img)
        # convert colors from BGR to RGB
        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
        # write the frame
        out.write(frame)
        # show the frame
        cv2.imshow("screenshot", frame)
        # if the user clicks q, it exits
        if cv2.waitKey(1) == ord("q"):
            break
    out.release()
    cv2.destroyAllWindows()


"""
This Script will help in Training the YOLOv3 Model on the custom dataset

It will take Prepared Data as Argument.

MODIFIED FROM keras-yolo3 PACKAGE, https://github.com/qqwweee/keras-yolo3
Retrain the YOLO model for your own dataset.

Author: Tushar Goel

"""

"""
Retrain the YOLO model for your own dataset.
"""

import numpy as np
import keras.backend as K
from keras.layers import Input, Lambda
from keras.models import Model
from keras.optimizers import Adam
from keras.callbacks import TensorBoard, ModelCheckpoint, ReduceLROnPlateau, EarlyStopping
import os
from .yolo4_model import preprocess_true_boxes, yolo4_body, yolo4_loss
from .yolo4_utils import get_random_data


def Train_Yolo(working_directory,model_name = 'yolov4.h5',input_shape = (608,608),val_split = 0.1,batch_size1 = 32,batch_size2 = 4,epochs1 = 51,epochs2 = 30,
               process1 = True,process2 = True):
    annotation_path = os.path.join(working_directory,'data_train.txt')
    log_dir = os.path.join(working_directory,'logs')
    if not os.path.exists(log_dir):
        os.mkdir(log_dir)
    classes_path = os.path.join(working_directory,'data_classes.txt')
    anchors_path = os.path.join(os.path.dirname(__file__),'yolo4_anchors.txt')
    class_names = get_classes(classes_path)
    num_classes = len(class_names)
    anchors = get_anchors(anchors_path)
    weights_path = os.path.join(working_directory,model_name)
    input_shape = input_shape # multiple of 32, hw

    model = create_model(input_shape, anchors, num_classes,
        freeze_body=2, weights_path=weights_path) # make sure you know what you freeze

    logging = TensorBoard(log_dir=log_dir)
    checkpoint = ModelCheckpoint(log_dir + 'ep{epoch:03d}-loss{loss:.3f}-val_loss{val_loss:.3f}.h5',
        monitor='val_loss', save_weights_only=True, save_best_only=True, period=1)
    reduce_lr = ReduceLROnPlateau(monitor='val_loss', factor=0.1, patience=3, verbose=1)
    early_stopping = EarlyStopping(monitor='val_loss', min_delta=0, patience=10, verbose=1)

    val_split = val_split
    with open(annotation_path) as f:
        lines = f.readlines()
    np.random.seed(10101)
    np.random.shuffle(lines)
    np.random.seed(None)
    num_val = int(len(lines)*val_split)
    num_train = len(lines) - num_val

    # Train with frozen layers first, to get a stable loss.
    # Adjust num epochs to your dataset. This step is enough to obtain a not bad model.
    if process1:
        model.compile(optimizer=Adam(lr=1e-3), loss={
            # use custom yolo4_loss Lambda layer.
            'yolo_loss': lambda y_true, y_pred: y_pred})

        batch_size = batch_size1
        print('Train on {} samples, val on {} samples, with batch size {}.'.format(num_train, num_val, batch_size))
        model.fit_generator(data_generator_wrapper(lines[:num_train], batch_size, input_shape, anchors, num_classes),
                steps_per_epoch=max(1, num_train//batch_size),
                validation_data=data_generator_wrapper(lines[num_train:], batch_size, input_shape, anchors, num_classes),
                validation_steps=max(1, num_val//batch_size),
                epochs=epochs1,
                initial_epoch=0,
                callbacks=[logging, checkpoint])
        model.save(log_dir + 'trained_weights_stage_1.h5')

    # Unfreeze and continue training, to fine-tune.
    # Train longer if the result is not good.
    if process2:
        for i in range(len(model.layers)):
            model.layers[i].trainable = True
        model.compile(optimizer=Adam(lr=1e-4), loss={'yolo_loss': lambda y_true, y_pred: y_pred}) # recompile to apply the change
        print('Unfreeze all of the layers.')

        batch_size = batch_size2 # note that more GPU memory is required after unfreezing the body
        print('Train on {} samples, val on {} samples, with batch size {}.'.format(num_train, num_val, batch_size))
        model.fit_generator(data_generator_wrapper(lines[:num_train], batch_size, input_shape, anchors, num_classes),
            steps_per_epoch=max(1, num_train//batch_size),
            validation_data=data_generator_wrapper(lines[num_train:], batch_size, input_shape, anchors, num_classes),
            validation_steps=max(1, num_val//batch_size),
            epochs=epochs2+epochs1,
            initial_epoch=epochs1,
            callbacks=[logging, checkpoint, reduce_lr, early_stopping])
        model.save(log_dir + 'trained_weights_final.h5')

    # Further training if needed.


def get_classes(classes_path):
    '''loads the classes'''
    with open(classes_path) as f:
        class_names = f.readlines()
    class_names = [c.strip() for c in class_names]
    return class_names

def get_anchors(anchors_path):
    '''loads the anchors from a file'''
    with open(anchors_path) as f:
        anchors = f.readline()
    anchors = [float(x) for x in anchors.split(',')]
    return np.array(anchors).reshape(-1, 2)


def create_model(input_shape, anchors, num_classes, load_pretrained=True, freeze_body=2,
            weights_path='model_data/yolo_weights.h5'):
    '''create the training model'''
    K.clear_session() # get a new session
    image_input = Input(shape=(None, None, 3))
    h, w = input_shape
    num_anchors = len(anchors)

    y_true = [Input(shape=(h//{0:32, 1:16, 2:8}[l], w//{0:32, 1:16, 2:8}[l], \
        num_anchors//3, num_classes+5)) for l in range(3)]

    model_body = yolo4_body(image_input, num_anchors//3, num_classes)
    print('Create YOLOv4 model with {} anchors and {} classes.'.format(num_anchors, num_classes))

    if load_pretrained:
        model_body.load_weights(weights_path, by_name=True, skip_mismatch=True)
        print('Load weights {}.'.format(weights_path))
        if freeze_body in [1, 2]:
            # Freeze darknet53 body or freeze all but 3 output layers.
            num = (250, len(model_body.layers)-3)[freeze_body-1]
            for i in range(num): model_body.layers[i].trainable = False
            print('Freeze the first {} layers of total {} layers.'.format(num, len(model_body.layers)))

    label_smoothing = 0
    use_focal_obj_loss = False
    use_focal_loss = False
    use_diou_loss = True
    use_softmax_loss = False

    model_loss = Lambda(yolo4_loss, output_shape=(1,), name='yolo_loss',
        arguments={'anchors': anchors, 'num_classes': num_classes, 'ignore_thresh': 0.5,
        'label_smoothing': label_smoothing, 'use_focal_obj_loss': use_focal_obj_loss, 'use_focal_loss': use_focal_loss, 'use_diou_loss': use_diou_loss,
        'use_softmax_loss': use_softmax_loss})(
        [*model_body.output, *y_true])
    model = Model([model_body.input, *y_true], model_loss)

    return model

def data_generator(annotation_lines, batch_size, input_shape, anchors, num_classes):
    '''data generator for fit_generator'''
    n = len(annotation_lines)
    i = 0
    while True:
        image_data = []
        box_data = []
        for b in range(batch_size):
            if i==0:
                np.random.shuffle(annotation_lines)
            image, box = get_random_data(annotation_lines[i], input_shape, random=True)
            image_data.append(image)
            box_data.append(box)
            i = (i+1) % n
        image_data = np.array(image_data)
        box_data = np.array(box_data)
        y_true = preprocess_true_boxes(box_data, input_shape, anchors, num_classes)
        yield [image_data, *y_true], np.zeros(batch_size)

def data_generator_wrapper(annotation_lines, batch_size, input_shape, anchors, num_classes):
    n = len(annotation_lines)
    if n==0 or batch_size<=0: return None
    return data_generator(annotation_lines, batch_size, input_shape, anchors, num_classes)



"""

This File will contain the implementation of YOLOV4 Algorithm,

containing Image_Annotation, Training and Prediction

Author: Tushar Goel

"""
import os
import wget
from .realtime_detect import Real_Time_Tracking
from .Yolo_Format import Image_Annotation
from .Download_Convert_Yolo_Weights import Yolo4_weights,Download_weights
from .Model_Training import Train_Yolo
from .Detection import Detector
from .web_cam_detect import Web_Cam_Detection
from .deep_sort_tracking import DeepSort_Tracking
from .yolo_tracking import YOLO_Tracker
import matplotlib.pyplot as plt

class YOLOv4:
    """
    This Class Will provide full implementation of YOLOv4 
    
    within simple steps.
    
    Attributes:
        working_directory --> The Directory where all Images and Labels have been provided
        
    Method:
        Prepare_the_Data --> This Method Will be used to Convert your raw into YOLO Format.
        
        Train_the_Yolo --> This Will Train the YOLOv3 For Users with no additional arguments
        
        Detect --> It will help in Detection of objects whether it is Image or Video.
        
    """
    
    def __init__(self,working_directory):
        
        # Creating Attribute Working Directory
        self.working_directory = working_directory
        print('\t\t###########################\n')
        print('\t\tWelcome To The DARK NEURON\n')
        print('\t\t###########################\n')
              
        coco_classes_path = os.path.join(os.path.dirname(__file__),'coco_classes.txt')
        anchors_path = os.path.join(os.path.dirname(__file__),'yolo4_anchors.txt')
        font_path = os.path.join(os.path.dirname(__file__),'FiraMono-Medium.otf')
        tracking_file_path = os.path.join(os.path.dirname(__file__),'mars-small128.pb')
        
        if not os.path.exists(coco_classes_path):
            wget.download('https://github.com/Tushar-ml/DarkNeuron/releases/download/v1.3.1/coco_classes.txt',coco_classes_path)
        
        if not os.path.exists(anchors_path):
            wget.download('https://github.com/Tushar-ml/DarkNeuron/releases/download/v1.3.1/yolo4_anchors.txt',anchors_path)
            
        if not os.path.exists(font_path):
            wget.download('https://github.com/Tushar-ml/DarkNeuron/releases/download/v1.3.1/FiraMono-Medium.otf',font_path)
            
        if not os.path.exists(tracking_file_path):
            wget.download('https://github.com/Tushar-ml/DarkNeuron/releases/download/v1.3.1/mars-small128.pb',tracking_file_path)
    
    def Prepare_the_Data(self,file_type,dataframe_name=None,class_file_name=None):
        """
        This Function will Prepare the Data according to the format of the file
        
        Arguments:
            
            file_type --> This Function will take dataframe,text or xml format
            dataframe_name --> Name of the Dataframe name in working directory. Only Needed when method = dataframe.
            class_file_name --> class_file_name containing label file names in .txt format
            
        Returns:
            None
        
        Additional Files Formed:
            data_train.txt --> Yolo Format of the Data
            data_classes.txt --> Classes of the Data
            
        """
        print('\n\t\t---------Preparing Data Phase-----------')   
        self.file_type = file_type
        # Method Involves 'csv','xml','text'
        Image_Annot = Image_Annotation(working_directory = self.working_directory,dataframe_name=dataframe_name)
        if self.file_type not in ['csv','xml','text']:
            raise ValueError('Method either should be dataframe,xml,text')
            
        if self.file_type == 'csv':
            print("Generating Data......")
            Image_Annot.Convert_to_Yolo_Format()
            
        elif self.file_type == 'xml':
            print("Generating Data......")
            print('Generating txt files from xml')
            Image_Annot.csv_from_xml(class_list_file_name = class_file_name)
            print('Generated')
            df = Image_Annot.csv_from_text(class_list_file_name = class_file_name)
           
            Image_Annot.Convert_to_Yolo_Format(df=df)

        elif self.file_type == 'text':
            if class_file_name is None:
                raise ValueError('Provide Class File Name for method text')
            print("Generating Data......")
            df = Image_Annot.csv_from_text(class_list_file_name = class_file_name)
            Image_Annot.Convert_to_Yolo_Format(df=df)
            
    def Train_the_Yolo(self,model_name = 'yolov4.h5',input_shape = (608,608),plot_model=False,save_weights=False,score = 0.5,iou = 0.5,
                       epochs1 = 51,epochs2 = 50, batch_size1 = 32,batch_size2 = 4,gpu_num = 1,
                       validation_split = 0.1,process1 = True,process2 = True):
        
        
        print('\n\t\t------------Training Phase Generated---------')
        
        yolo_file_path = os.path.join(self.working_directory,'yolov4.h5')
        self.class_path = os.path.join(self.working_directory,'class_list.txt')
        self.anchors_path = os.path.join(os.path.dirname(__file__),'yolo4_anchors.txt')
        self.weight_path = os.path.join(self.working_directory,'yolov4.weights')
        self.coco_class = os.path.join(os.path.dirname(__file__),'coco_classes.txt')
        #Checking whether User have Yolo File or Not 
        #If no File, then it will be downloaded Automatically and Converted to Keras Model
        if not os.path.exists(yolo_file_path):
            
            Download_weights(working_directory = self.working_directory)
            yolov4 = Yolo4_weights(score=score,iou=iou,anchors_path = self.anchors_path,classes_path = self.coco_class,
                                   model_path = yolo_file_path,weights_path = self.weight_path,gpu_num = gpu_num)
            yolov4.load_yolo()

        
        print('Model Training to be Start ....')
        history = Train_Yolo(working_directory = self.working_directory,model_name = model_name,val_split = validation_split,
                   epochs1=epochs1,epochs2 = epochs2,batch_size1 = batch_size1,
                   batch_size2 = batch_size2,process1 = process1, process2 = process2)
        
        
        self.history = history
        
        try:
            plt.plot(self.history.history['loss'],label = 'Training Loss')
            plt.plot(self.history.history['val_loss'],label = 'Validation Loss')
            plt.title('Training Loss vs Validation Loss')
            plt.legends()
            plt.show()
        
        except:
            pass
        
    def Detect(self,test_folder_name='test',model_name = None,cam = False,real_time = False,videopath = 0,classes = [],tracking =False,score=0.5,gpu_num = 1):
        
        if model_name is None:
            model_path = os.path.join(self.working_directory,'yolov4.h5')
            print('Yolo File Not Found...Downloading and Converting...')
            if not os.path.exists(model_path):
                Download_weights(working_directory = self.working_directory)
                yolov4 = Yolo4_weights(score=score,anchors_path = self.anchors_path,classes_path = self.coco_class,
                                   model_path = model_path,weights_path = self.weight_path,gpu_num = gpu_num)
                yolov4.load_yolo()
        
        yolo_tracker = YOLO_Tracker(working_directory = self.working_directory,model_name=model_name,
                                   score = score,classes = classes)
        
        print('\n\t\t-------------Detection Phase Generated-------------')
        
        if real_time:
            Real_Time_Tracking(working_directory = self.working_directory,classes = classes,score = score)
        
        elif cam:

            Web_Cam_Detection(working_directory = self.working_directory,videopath = videopath,model_name=model_name,score=score,
                              gpu_num = gpu_num,classes = classes)
            
        elif tracking :
                
            DeepSort_Tracking(working_directory = self.working_directory,yolo = yolo_tracker,file_path = videopath)
        
        else:
            Detector(working_directory = self.working_directory,test_folder_name = test_folder_name,
                     classes = classes,score = score,model_name = model_name,gpu_num = gpu_num)
            
    
       




from __future__ import division, print_function, absolute_import

from timeit import time
import warnings
import cv2
import numpy as np
from PIL import Image
import os
from .deepsort_preprocessing import *
from .nn_matching import *
from .deepsort_detection import Detection
from .tracker import Tracker
from  .generate_detections import *
import imutils.video


warnings.filterwarnings('ignore')

def DeepSort_Tracking(yolo,working_directory,file_path = 0):

    # Definition of the parameters
    max_cosine_distance = 0.3
    nn_budget = None
    nms_max_overlap = 1.0
    
    # Deep SORT
    model_filename = os.path.join(os.path.dirname(__file__),'mars-small128.pb')
    print('Deep Sort Model Generated')
    encoder = create_box_encoder(model_filename,batch_size=1)
    
    metric = NearestNeighborDistanceMetric("cosine", max_cosine_distance, nn_budget)
    tracker = Tracker(metric)

    writeVideo_flag = True
    

    file_path = file_path
    video_capture = cv2.VideoCapture(file_path)



    if writeVideo_flag:
        w = int(video_capture.get(3))
        h = int(video_capture.get(4))
        fourcc = cv2.VideoWriter_fourcc(*'XVID')
        out = cv2.VideoWriter(os.path.join(working_directory,'output_Tracking.avi'), fourcc, 30, (w, h))
        frame_index = -1

    fps = 0.0
    fps_imutils = imutils.video.FPS().start()

    while True:
        ret, frame = video_capture.read()  # frame shape 640*480*3
        if ret != True:
             break

        t1 = time.time()

        image = Image.fromarray(frame[...,::-1])  # bgr to rgb
        boxs = yolo.detect_image(image)[0]
        confidence = yolo.detect_image(image)[1]

        features = encoder(frame,boxs)

        detections = [Detection(bbox, confidence, feature) for bbox, confidence, feature in zip(boxs, confidence, features)]
        
        # Run non-maxima suppression.
        boxes = np.array([d.tlwh for d in detections])
        scores = np.array([d.confidence for d in detections])
        indices = non_max_suppression(boxes, nms_max_overlap, scores)
        detections = [detections[i] for i in indices]
        
        # Call the tracker
        tracker.predict()
        tracker.update(detections)
        
        for track in tracker.tracks:
            if not track.is_confirmed() or track.time_since_update > 1:
                continue 
            bbox = track.to_tlbr()
            cv2.rectangle(frame, (int(bbox[0]), int(bbox[1])), (int(bbox[2]), int(bbox[3])),(255,255,255), 2)
            cv2.putText(frame, str(track.track_id),(int(bbox[0]), int(bbox[1])),0, 5e-3 * 200, (0,255,0),2)

        for det in detections:
            bbox = det.to_tlbr()
            score = "%.2f" % round(det.confidence * 100, 2)
            cv2.rectangle(frame,(int(bbox[0]), int(bbox[1])), (int(bbox[2]), int(bbox[3])),(255,0,0), 2)
            cv2.putText(frame, score + '%', (int(bbox[0]), int(bbox[3])), 0, 5e-3 * 130, (0,255,0),2)
            
        cv2.imshow('', frame)

        if writeVideo_flag: # and not asyncVideo_flag:
            # save a frame
            out.write(frame)
            frame_index = frame_index + 1

        fps_imutils.update()

        fps = (fps + (1./(time.time()-t1))) / 2
        print("FPS = %f"%(fps))
        
        # Press Q to stop!
        if cv2.waitKey(1) & 0xFF == ord('q'):
            break

    fps_imutils.stop()
    print('imutils FPS: {}'.format(fps_imutils.fps()))
    video_capture.release()

    if writeVideo_flag:
        out.release()

    cv2.destroyAllWindows()




# -*- coding: utf-8 -*-
"""
Created on Tue Jun  2 19:48:33 2020

@author: Tushar Goel
"""

import glob
import os
import pickle
import xml.etree.ElementTree as ET
from os import listdir, getcwd
from os.path import join
import cv2

def getImagesInDir(dirName, endings=[".jpg", ".jpeg", ".png", ".mp4"]):
        # create a list of file and sub directories
        # names in the given directory
        listOfFile = os.listdir(dirName)
        allFiles = list()
        # Make sure all file endings start with a '.'
        endings_final = [0]*len(endings)
        for i, ending in enumerate(endings):
            if ending[0] != ".":
                endings_final[i] = "." + ending
        # Iterate over all the entries
        for entry in listOfFile:
            # Create full path
            fullPath = os.path.join(dirName, entry)
            # If entry is a directory then get the list of files in this directory
            if os.path.isdir(fullPath):
                allFiles = allFiles + getImagesInDir(fullPath, endings)
            else:
                for ending in endings:
                    if entry.endswith(ending):
                        allFiles.append(fullPath)
        return allFiles

def convert(size, box):
    dw = 1./size[0]
    dh = 1./size[1]

    xmin = box[0]*dw
    ymin = box[2]*dh
    xmax = box[1]*dw
    ymax = box[3]*dh
    
    return (xmin,ymin,xmax,ymax)

def convert_annotation(dir_path, output_path, image_path,classes):
    basename = os.path.basename(image_path)
    basename_no_ext = os.path.splitext(basename)[0]
    img = cv2.imread(image_path)
    
    in_file = open(dir_path + '\\' + basename_no_ext + '.xml')
    
    out_file = open(output_path +'\\'+ basename_no_ext + '.txt', 'w')
    tree = ET.parse(in_file)
    root = tree.getroot()
    size = root.find('size')
    w = img.shape[1]
    h = img.shape[0]
    for obj in root.iter('object'):
        difficult = obj.find('difficult').text
        
        cls = obj.find('name').text.lower()
        if cls not in classes or int(difficult)==1:
            continue
        cls_id = classes.index(cls)
        
        xmlbox = obj.find('bndbox')
        
        b = (float(xmlbox.find('xmin').text), float(xmlbox.find('xmax').text), float(xmlbox.find('ymin').text), float(xmlbox.find('ymax').text))
        bb = convert((w,h),b)
        out_file.write(str(cls_id) + " " + " ".join([str(a) for a in bb]) + '\n')
    out_file.close()


# -*- coding: utf-8 -*-
"""
Class definition of YOLO_v3 style detection model on image and video
"""

import colorsys
import os
from timeit import default_timer as timer
import tensorflow as tf
import numpy as np
from keras import backend as K
from keras.models import load_model
from keras.layers import Input
from PIL import Image, ImageFont, ImageDraw

from .yolo4_model import yolo_eval, yolo4_body,tiny_yolo_body
from .yolo4_utils import letterbox_image
from keras.utils import multi_gpu_model

class YOLO:
    _defaults = {
        "model_path": 'model_data/yolo4.h5',
        "anchors_path": os.path.join(os.path.dirname(__file__),'yolo4_anchors.txt'),
        "classes_path": os.path.join(os.path.dirname(__file__),'coco_classes.txt'),
        "score" : 0.5,
        "iou" : 0.45,
        "model_image_size" : (608, 608),
        "gpu_num" : 1,
    }

    @classmethod
    def get_defaults(cls, n):
        if n in cls._defaults:
            return cls._defaults[n]
        else:
            return "Unrecognized attribute name '" + n + "'"

    def __init__(self, **kwargs):
        self.__dict__.update(self._defaults) # set up default values
        self.__dict__.update(kwargs) # and update with user overrides
        self.class_names = self._get_class()
        self.anchors = self._get_anchors()
        self.sess = K.get_session()
        self.boxes, self.scores, self.classes = self.generate()

    def _get_class(self):
        classes_path = os.path.expanduser(self.classes_path)
        with open(classes_path) as f:
            class_names = f.readlines()
        class_names = [c.strip() for c in class_names]
        return class_names

    def _get_anchors(self):
        anchors_path = os.path.expanduser(self.anchors_path)
        with open(anchors_path) as f:
            anchors = f.readline()
        anchors = [float(x) for x in anchors.split(',')]
        return np.array(anchors).reshape(-1, 2)

    def generate(self):
        model_path = os.path.expanduser(self.model_path)
        assert model_path.endswith(".h5"), "Keras model or weights must be a .h5 file."

        # Load model, or construct model and load weights.
        start = timer()
        num_anchors = len(self.anchors)
        num_classes = len(self.class_names)
        is_tiny_version = num_anchors == 6  # default setting
        try:
            self.yolo_model = load_model(model_path, compile=False)
        except:
            self.yolo_model = (
                tiny_yolo_body(
                    Input(shape=(None, None, 3)), num_anchors // 2, num_classes
                )
                if is_tiny_version
                else yolo4_body(
                    Input(shape=(None, None, 3)), num_anchors // 3, num_classes
                )
            )
            self.yolo_model.load_weights(
                self.model_path
            )  # make sure model, anchors and classes match
        else:
            assert self.yolo_model.layers[-1].output_shape[-1] == num_anchors / len(
                self.yolo_model.output
            ) * (
                num_classes + 5
            ), "Mismatch between model and given anchor and class sizes"

        end = timer()
        print(
            "{} model, anchors, and classes loaded in {:.2f}sec.".format(
                model_path, end - start
            )
        )

        # Generate colors for drawing bounding boxes.
        if len(self.class_names) == 1:
            self.colors = ["GreenYellow"]
        else:
            hsv_tuples = [
                (x / len(self.class_names), 1.0, 1.0)
                for x in range(len(self.class_names))
            ]
            self.colors = list(map(lambda x: colorsys.hsv_to_rgb(*x), hsv_tuples))
            self.colors = list(
                map(
                    lambda x: (int(x[0] * 255), int(x[1] * 255), int(x[2] * 255)),
                    self.colors,
                )
            )
            np.random.seed(10101)  # Fixed seed for consistent colors across runs.
            np.random.shuffle(
                self.colors
            )  # Shuffle colors to decorrelate adjacent classes.
            np.random.seed(None)  # Reset seed to default.

        # Generate output tensor targets for filtered bounding boxes.
        self.input_image_shape = K.placeholder(shape=(2,))
        if self.gpu_num >= 2:
            self.yolo_model = multi_gpu_model(self.yolo_model, gpus=self.gpu_num)
        boxes, scores, classes = yolo_eval(
            self.yolo_model.output,
            self.anchors,
            len(self.class_names),
            self.input_image_shape,
            score_threshold=self.score,
            iou_threshold=self.iou,
        )
        return boxes, scores, classes

    def detect_image(self, image,classes = [],score = 0.5,show_stats=True):
        start = timer()
        # self.sess.run(tf.compat.v1.global_variables_initializer())
        if self.model_image_size != (None, None):
            assert self.model_image_size[0] % 32 == 0, "Multiples of 32 required"
            assert self.model_image_size[1] % 32 == 0, "Multiples of 32 required"
            boxed_image = letterbox_image(image, tuple(reversed(self.model_image_size)))
        else:
            new_image_size = (
                image.width - (image.width % 32),
                image.height - (image.height % 32),
            )
            boxed_image = letterbox_image(image, new_image_size)
        image_data = np.array(boxed_image, dtype="float32")
        if show_stats:
            print(image_data.shape)
        image_data /= 255.0
        image_data = np.expand_dims(image_data, 0)  # Add batch dimension.

        out_boxes, out_scores, out_classes = self.sess.run(
            [self.boxes, self.scores, self.classes],
            feed_dict={
                self.yolo_model.input: image_data,
                self.input_image_shape: [image.size[1], image.size[0]],
                K.learning_phase(): 0,
            },
        )
        
        if show_stats:
            print("Found {} boxes for {}".format(len(out_boxes), "img"))
        out_prediction = []

        font_path = os.path.join(os.path.dirname(__file__), "FiraMono-Medium.otf")
        font = ImageFont.truetype(
            font=font_path, size=np.floor(3e-2 * image.size[1] + 0.5).astype("int32")
        )
        thickness = (image.size[0] + image.size[1]) // 300
        if type(classes) is not list:
            raise SyntaxError('Provide the Classes as List')
            
            
        for i, c in reversed(list(enumerate(out_classes))):
            if (len(classes) != 0):
                if self.class_names[c] in classes:
                    predicted_class = self.class_names[c]
                else:
                    continue
            else:
                predicted_class = self.class_names[c]
            box = out_boxes[i]
            if out_scores[i]>score:
                score = out_scores[i]
                
            else:
                continue

            label = "{} {:.2f}".format(predicted_class, score)
            draw = ImageDraw.Draw(image)
            label_size = draw.textsize(label, font)

            top, left, bottom, right = box
            top = max(0, np.floor(top + 0.5).astype("int32"))
            left = max(0, np.floor(left + 0.5).astype("int32"))
            bottom = min(image.size[1], np.floor(bottom + 0.5).astype("int32"))
            right = min(image.size[0], np.floor(right + 0.5).astype("int32"))

            # image was expanded to model_image_size: make sure it did not pick
            # up any box outside of original image (run into this bug when
            # lowering confidence threshold to 0.01)
            if top > image.size[1] or right > image.size[0]:
                continue
            if show_stats:
                print(label, (left, top), (right, bottom))

            # output as xmin, ymin, xmax, ymax, class_index, confidence
            out_prediction.append([left, top, right, bottom, c, score])

            if top - label_size[1] >= 0:
                text_origin = np.array([left, top - label_size[1]])
            else:
                text_origin = np.array([left, bottom])

            # My kingdom for a good redistributable image drawing library.
            for i in range(thickness):
                draw.rectangle(
                    [left + i, top + i, right - i, bottom - i], outline=self.colors[c]
                )
            draw.rectangle(
                [tuple(text_origin), tuple(text_origin + label_size)],
                fill=self.colors[c],
            )

            draw.text(text_origin, label, fill=(0, 0, 0), font=font)
            del draw

        end = timer()
        if show_stats:
            print("Time spent: {:.3f}sec".format(end - start))
        
        return out_prediction, image
        self.close_session()

    def close_session(self):
        self.sess.close()

def detect_video(yolo, video_path, classes = [],score = 0.5,output_path=""):
    import cv2
    vid = cv2.VideoCapture(video_path)
    if not vid.isOpened():
        raise IOError("Couldn't open webcam or video")
    video_FourCC    = int(vid.get(cv2.CAP_PROP_FOURCC))
    video_fps       = vid.get(cv2.CAP_PROP_FPS)
    video_size      = (int(vid.get(cv2.CAP_PROP_FRAME_WIDTH)),
                        int(vid.get(cv2.CAP_PROP_FRAME_HEIGHT)))
    isOutput = True if output_path != "" else False
    if isOutput:
        print("!!! TYPE:", type(output_path), type(video_FourCC), type(video_fps), type(video_size))
        out = cv2.VideoWriter(output_path, video_FourCC, video_fps, video_size)
    accum_time = 0
    curr_fps = 0
    fps = "FPS: ??"
    prev_time = timer()
    while True:
        return_value, frame = vid.read()
        image = Image.fromarray(frame)
        prediction,image = yolo.detect_image(image,classes=classes,score=score,show_stats=False)
        result = np.asarray(image)
        curr_time = timer()
        exec_time = curr_time - prev_time
        prev_time = curr_time
        accum_time = accum_time + exec_time
        curr_fps = curr_fps + 1
        if accum_time > 1:
            accum_time = accum_time - 1
            fps = "FPS: " + str(curr_fps)
            curr_fps = 0
        cv2.putText(result, text=fps, org=(3, 15), fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                    fontScale=0.50, color=(255, 0, 0), thickness=2)
        cv2.namedWindow("result", cv2.WINDOW_NORMAL)
        cv2.imshow("result", result)
        if isOutput:
            out.write(result)
        if cv2.waitKey(1) & 0xFF == ord('q'):
            break
    yolo.close_session()




"""
This Script will convert Images and Labels in the Yolo Format

Images and Labels Can be In csv Format, will be guided in Readme.md File

about Microsoft VOTT (Visual Object Tagging Tool) and How to export csv File 

from it. 

Author : Tushar Goel

"""
import cv2
from PIL import Image
import os
import re
import pandas as pd
import numpy as np
from glob import glob
from .voc_to_yolo import convert_annotation, getImagesInDir

class Image_Annotation:
    """
    This Class will used for conversion of Image and labels into YOLO format
    
    Attributes:
        working_directory --> working_directory where files and Labels will be there.
        Dataframe Name --> DataFrame Name containing Label Files and Images
        
    Methods:
        Convert_to_Yolo_Format
        Convert_csv_to_Yolo
        csv_from_xml
        csv_from_text
        
    Additional Files Formed:
        data_train.txt --> Containing Images and Labels in YOLO Format
        data_classes.txt --> having Classes of Custom Object
        
    """
    
    def __init__(self,working_directory,dataframe_name=None):
        
        self.working_directory = working_directory
        
        #Initialing Variables Required in Image Annotation
        """
        self.data_train --> Containing File and labels in Yolo Format
        self.data_classes --> Containing Classes of the data
        
        """
        if dataframe_name is not None:
            self.dataframe_path = os.path.join(self.working_directory,dataframe_name)
        self.data_train = os.path.join(self.working_directory,'data_train.txt')
        self.data_classes = os.path.join(self.working_directory,'data_classes.txt')
        
    
    def Convert_to_Yolo_Format(self,df=None):
        
        if df is not None:
            multi_df = df
        else:
            multi_df = pd.read_csv(self.dataframe_path)
        labels = multi_df["label"].unique()
        labeldict = dict(zip(labels, range(len(labels))))
        multi_df.drop_duplicates(subset=None, keep="first", inplace=True)
        train_path = self.working_directory
        self.convert_csv_to_yolo(
            multi_df, labeldict, path=train_path, target_name=self.data_train
        )
    
        # Make classes file
        file = open(self.data_classes, "w")
    
        # Sort Dict by Values
        SortedLabelDict = sorted(labeldict.items(), key=lambda x: x[1])
        for elem in SortedLabelDict:
            file.write(elem[0] + "\n")
        file.close()
        
    def convert_csv_to_yolo(   self,
                                    vott_df,
                                    labeldict=dict(zip(["Yolo_Training"], [0,])),
                                    path="",
                                    target_name="data_train.txt",
                                    abs_path=False,
                                ):
        
    # Encode labels according to labeldict if code's don't exist
        if not "code" in vott_df.columns:
            vott_df["code"] = vott_df["label"].apply(lambda x: labeldict[x])
        # Round float to ints
        for col in vott_df[["xmin", "ymin", "xmax", "ymax"]]:
            vott_df[col] = (vott_df[col]).apply(lambda x: round(x))
    
        # Create Yolo Text file
        last_image = ""
        txt_file = ""
    
        for index, row in vott_df.iterrows():
            if not last_image == row["image"]:
                if abs_path:
                    txt_file += "\n" + row["image_path"] + " "
                else:
                    txt_file += "\n" + os.path.join(path, row["image"]) + " "
                txt_file += ",".join(
                    [
                        str(x)
                        for x in (row[["xmin", "ymin", "xmax", "ymax", "code"]].tolist())
                    ]
                )
            else:
                txt_file += " "
                txt_file += ",".join(
                    [
                        str(x)
                        for x in (row[["xmin", "ymin", "xmax", "ymax", "code"]].tolist())
                    ]
                )
            last_image = row["image"]
        file = open(target_name, "w")
        file.write(txt_file[1:])
        file.close()
        return True
    
    def csv_from_xml(self,class_list_file_name):
        output_path = self.working_directory
        class_file_text = os.path.join(self.working_directory,class_list_file_name)
        class_file = open(class_file_text,'r')
        classes = class_file.readlines()

        class_list = []
        for cla in classes:
            cla = cla.split()[0]
            class_list.append(cla.lower())
            
        class_list = sorted(class_list)
        image_paths = getImagesInDir(output_path)
        list_file = open(output_path + '.txt', 'w')
        for image_path in image_paths:
            list_file.write(image_path + '\n')
            convert_annotation(output_path, output_path, image_path,class_list)
        list_file.close()
        
    def csv_from_text(self,class_list_file_name):
        
        text_file_paths = glob(os.path.join(self.working_directory,'*.txt'))
        image_file_paths = GetFileList(self.working_directory,['.jpg','.jpeg','.png'])
        #Removing Class_text file from text path:
        
        class_file_text = os.path.join(self.working_directory,class_list_file_name)
        text_file_paths.remove(class_file_text) #This is not our Data 
        assert len(text_file_paths) == len(image_file_paths),"Length of image files and Their corressponding text files does not match"
   
        class_file = open(class_file_text,'r')
        classes = class_file.readlines()
        class_dic = dict()
        count = 0
        class_list = []
        for cla in classes:
            cla = cla.split()[0]
            class_list.append(cla)
            
        class_list = sorted(class_list)
        for cla in class_list:
            cla = cla.split()[0]
            class_dic[count] = cla
            count += 1
        image_name = []
        xmin = []
        ymin = []
        xmax = []
        ymax = []
        label = []
        df = pd.DataFrame()
        
        for text_file in text_file_paths:
            for i in image_file_paths:
                try:
                    if text_file.split('\\')[-1].split('.')[0] == i.split('\\')[-1].split('.')[0]:
                        image_path = i
                        break
                except:
                    if text_file.split('/')[-1].split('.')[0] == i.split('/')[-1].split('.')[0]:
                        image_path = i
                        break
            #y_size, x_size = np.array(Image.open(image_path)).shape
            file = open(text_file,'r')
            img = cv2.imread(image_path)
            w = img.shape[1]
            h = img.shape[0]
            lines = file.readlines()
            for line in lines:
                line = line.split()
                image_name.append(image_path)
                xmin.append(float(line[1])*w)
                ymin.append(float(line[2])*h)
                xmax.append(float(line[3])*w)
                ymax.append(float(line[4])*h)
                
                label.append(class_dic[int(line[0])])
                
        df['image'] = image_name
        df['xmin'] = xmin
        df['ymin'] = ymin
        df['xmax'] = xmax
        df['ymax'] = ymax
        df['label'] = label

        return df
        
    
    
def GetFileList(dirName, endings=[".jpg", ".jpeg", ".png", ".mp4"]):
        # create a list of file and sub directories
        # names in the given directory
        listOfFile = os.listdir(dirName)
        allFiles = list()
        # Make sure all file endings start with a '.'
        endings_final = [0]*len(endings)
        for i, ending in enumerate(endings):
            if ending[0] != ".":
                endings_final[i] = "." + ending
        # Iterate over all the entries
        for entry in listOfFile:
            # Create full path
            fullPath = os.path.join(dirName, entry)
            # If entry is a directory then get the list of files in this directory
            if os.path.isdir(fullPath):
                allFiles = allFiles + GetFileList(fullPath, endings)
            else:
                for ending in endings:
                    if entry.endswith(ending):
                        allFiles.append(fullPath)
        return allFiles




""" This Package will deal with implementation of Automatic Deep Learning 

which can reduce the time and Complexity for non-technical users

to train their own netwroks with significant accuracies along with

Deployement with Tensorflow Serving 

Author: Tushar Goel

"""

class Dark_Neuron:
    """ This will be the Parent class for Deep Learning Techniques
    
    developed in this Package: Object Detection, Image Segmentation, Classfication 
    
    of Images, Certain Predefined Models """
    
    def __init__(self,working_directory=None,output_directory=None):
        """
        This Function will take the Directory of Data to be used
        
        and output directory for corresponding results___
        
        """
        self.working_directory= working_directory       # Input Directory
        self.output_directory = output_directory        # Output Directory
        
        print('\t\t###########################\n')
        print('\t\tWelcome To The DARK NEURON\n')
        print('\t\t###########################\n')
              
              
              



import colorsys
import cv2
import h5py
from keras import Model
import numpy as np
import os
from matplotlib.colors import rgb_to_hsv, hsv_to_rgb
from PIL import Image, ImageFont, ImageDraw
from timeit import default_timer as timer

# import readline
# readline.parse_and_bind("tab: complete")

min_logo_size = (10, 10)


def detect_object(yolo, img_path, save_img,classes = [],score = 0.5, save_img_path="./", postfix=""):
    """
    Call YOLO logo detector on input image, optionally save resulting image.

    Args:
      yolo: keras-yolo3 initialized YOLO instance
      img_path: path to image file
      save_img: bool to save annotated image
      save_img_path: path to directory where to save image
      postfix: string to add to filenames
    Returns:
      prediction: list of bounding boxes in format (xmin,ymin,xmax,ymax,class_id,confidence)
      image: unaltered input image as (H,W,C) array
    """
    try:
        image = Image.open(img_path)
        if image.mode != "RGB":
            image = image.convert("RGB")
        image_array = np.array(image)
    except:
        print("File Open Error! Try again!")
        return None, None

    prediction, new_image = yolo.detect_image(image,classes = classes,score=score)

    img_out = postfix.join(os.path.splitext(os.path.basename(img_path)))
    if save_img:
        new_image.save(os.path.join(save_img_path, img_out))

    return prediction, image_array


def parse_input():
    """
    Ask user input for input images: pass path to individual images, directory
    """
    out = []
    while True:
        ins = input("Enter path (q to quit):").strip()
        if ins in ["q", "quit"]:
            break
        if not os.path.exists(ins):
            print("Error: file not found!")
        elif os.path.isdir(ins):
            out = [
                os.path.abspath(os.path.join(ins, f))
                for f in os.listdir(ins)
                if f.endswith((".jpg", ".png"))
            ]
            break
        elif ins.endswith((".jpg", ".png")):
            out.append(os.path.abspath(ins))
        print(out)
    return out


def load_extractor_model(model_name="InceptionV3", flavor=1):
    """Load variant of InceptionV3 or VGG16 model specified.

    Args:
      model_name: string, either InceptionV3 or VGG16
      flavor: int specifying the model variant and input_shape.
        For InceptionV3, the map is {0: default, 1: 200*200, truncate last Inception block,
        2: 200*200, truncate last 2 blocks, 3: 200*200, truncate last 3 blocks, 4: 200*200}
        For VGG16, it only changes the input size, {0: 224 (default), 1: 128, 2: 64}.
"""
    start = timer()
    if model_name == "InceptionV3":
        from keras.applications.inception_v3 import InceptionV3
        from keras.applications.inception_v3 import preprocess_input

        model = InceptionV3(weights="imagenet", include_top=False)

        trunc_layer = [-1, 279, 248, 228, -1]
        i_layer = flavor
        model_out = Model(
            inputs=model.inputs, outputs=model.layers[trunc_layer[i_layer]].output
        )
        input_shape = (299, 299, 3) if flavor == 0 else (200, 200, 3)

    elif model_name == "VGG16":
        from keras.applications.vgg16 import VGG16
        from keras.applications.vgg16 import preprocess_input

        model_out = VGG16(weights="imagenet", include_top=False)
        input_length = [224, 128, 64][flavor]
        input_shape = (input_length, input_length, 3)

    end = timer()
    print("Loaded {} feature extractor in {:.2f}sec".format(model_name, end - start))
    return model_out, preprocess_input, input_shape


def chunks(l, n, preprocessing_function=None):
    """Yield successive n-sized chunks from l.

    General purpose function modified for Keras: made infinite loop,
    add preprocessing, returns np.array instead of list

    Args:
      l: iterable
      n: number of items to take for each chunk
      preprocessing_function: function that processes image (3D array)
    Returns:
      generator with n-sized np.array preprocessed chunks of the input
    """

    func = (lambda x: x) if (preprocessing_function is None) else preprocessing_function

    # in predict_generator, steps argument sets how many times looped through "while True"
    while True:
        for i in range(0, len(l), n):
            yield np.array([func(el) for el in l[i : i + n]])


def load_features(filename):
    """
    Load pre-saved HDF5 features for all logos in the LogosInTheWild database
    """

    start = timer()
    # get database features
    with h5py.File(filename, "r") as hf:
        brand_map = list(hf.get("brand_map"))
        input_shape = list(hf.get("input_shape"))
        features = hf.get("features")
        features = np.array(features)
    end = timer()
    print(
        "Loaded {} features from {} in {:.2f}sec".format(
            features.shape, filename, end - start
        )
    )

    return features, brand_map, input_shape


def save_features(filename, features, brand_map, input_shape):
    """
    Save features to compressed HDF5 file for later use
    """

    print("Saving {} features into {}... ".format(features.shape, filename), end="")
    # reduce file size by saving as float16
    features = features.astype(np.float16)
    start = timer()
    with h5py.File(filename, "w") as hf:
        hf.create_dataset("features", data=features, compression="lzf")
        hf.create_dataset("brand_map", data=brand_map)
        hf.create_dataset("input_shape", data=input_shape)

    end = timer()
    print("done in {:.2f}sec".format(end - start))

    return None


def features_from_image(img_array, model, preprocess, batch_size=100):
    """
    Extract features from image array given a decapitated keras model.
    Use a generator to avoid running out of memory for large inputs.

    Args:
      img_array: (N, H, W, C) list/array of input images
      model: keras model, outputs
    Returns:
      features: (N, F) array of 1D features
    """

    if len(img_array) == 0:
        return np.array([])

    steps = len(img_array) // batch_size + 1
    img_gen = chunks(img_array, batch_size, preprocessing_function=preprocess)
    features = model.predict_generator(img_gen, steps=steps)

    # if the generator has looped past end of array, cut it down
    features = features[: len(img_array)]

    # reshape features: flatten last three dimensions to one
    features = features.reshape(features.shape[0], np.prod(features.shape[1:]))
    return features


##################################################
# image processing and bounding box functions
##################################################


def pad_image(img, shape, mode="constant_mean"):
    """
    Resize and pad image to given size.

    Args:
      img: (H, W, C) input numpy array
      shape: (H', W') destination size
      mode: filling mode for new padded pixels. Default = 'constant_mean' returns
        grayscale padding with pixel intensity equal to mean of the array. Other
        options include np.pad() options, such as 'edge', 'mean' (by row/column)...
    Returns:
      new_im: (H', W', C) padded numpy array
    """
    if mode == "constant_mean":
        mode_args = {"mode": "constant", "constant_values": np.mean(img)}
    else:
        mode_args = {"mode": mode}

    ih, iw = img.shape[:2]
    h, w = shape[:2]

    # first rescale image so that largest dimension matches target
    scale = min(w / iw, h / ih)
    nw, nh = int(iw * scale), int(ih * scale)
    img = cv2.resize(img, (nw, nh))

    # center-pad rest of image: compute padding and split in two
    xpad, ypad = shape[1] - nw, shape[0] - nh
    xpad = (xpad // 2, xpad // 2 + xpad % 2)
    ypad = (ypad // 2, ypad // 2 + ypad % 2)

    new_im = np.pad(img, pad_width=(ypad, xpad, (0, 0)), **mode_args)

    return new_im


def bbox_colors(n):
    """
    Define n distinct bounding box colors

    Args:
      n: number of colors
    Returns:
      colors: (n, 3) np.array with RGB integer values in [0-255] range
    """
    hsv_tuples = [(x / n, 1.0, 1.0) for x in range(n)]
    colors = 255 * np.array([colorsys.hsv_to_rgb(*x) for x in hsv_tuples])

    np.random.seed(10101)  # Fixed seed for consistent colors across runs.
    np.random.shuffle(colors)  # Shuffle colors to decorrelate adjacent classes.
    np.random.seed(None)  # Reset seed to default.

    return colors.astype(int)


def contents_of_bbox(img, bbox_list, expand=1.0):
    """
    Extract portions of image inside  bounding boxes list.

    Args:
      img: (H,W,C) image array
      bbox_list: list of bounding box specifications, with first 4 elements
      specifying box corners in (xmin, ymin, xmax, ymax) format.
    Returns:
      candidates: list of 3D image arrays
      i_candidates_too_small: list of indices of small candidates dropped
    """

    candidates = []
    i_candidates_too_small = []
    for i, (xmin, ymin, xmax, ymax, *_) in enumerate(bbox_list):

        # for very low confidence sometimes logos found outside of the image
        if ymin > img.shape[0] or xmin > img.shape[1]:
            continue

        xmin, ymin = int(xmin // expand), int(ymin // expand)
        xmax, ymax = int(np.round(xmax // expand)), int(np.round(ymax // expand))

        # do not even consider tiny logos
        if xmax - xmin > min_logo_size[1] and ymax - ymin > min_logo_size[0]:
            candidates.append(img[ymin:ymax, xmin:xmax])
        else:
            i_candidates_too_small.append(i)

    return candidates, i_candidates_too_small


def draw_annotated_box(image, box_list_list, label_list, color_list):
    """
    Draw box and overhead label on image.

    Args:
      image: PIL image object
      box_list_list: list of lists of bounding boxes, one for each label, each box in
        (xmin, ymin, xmax, ymax [, score]) format (where score is an optional float)
      label_list: list of  string to go above box
      color_list: list of RGB tuples
    Returns:
      image: annotated PIL image object
    """

    font_path = "font/FiraMono-Medium.otf"

    font = ImageFont.truetype(
        font=font_path, size=np.floor(3e-2 * image.size[1] + 0.5).astype("int32")
    )
    thickness = (image.size[0] + image.size[1]) // 300

    draw = ImageDraw.Draw(image)

    for box_list, label, color in zip(box_list_list, label_list, color_list):
        if not isinstance(color, tuple):
            color = tuple(color)
        for box in box_list:
            # deal with empty predictions
            if len(box) < 4:
                continue

            # if score is also passed, append to label
            thelabel = "{}".format(label)
            if len(box) > 4:
                thelabel += " {:.2f}".format(box[-1])
            label_size = draw.textsize(thelabel, font)

            xmin, ymin, xmax, ymax = box[:4]
            ymin = max(0, np.floor(ymin + 0.5).astype("int32"))
            xmin = max(0, np.floor(xmin + 0.5).astype("int32"))
            ymax = min(image.size[1], np.floor(ymax + 0.5).astype("int32"))
            xmax = min(image.size[0], np.floor(xmax + 0.5).astype("int32"))

            if ymin - label_size[1] >= 0:
                text_origin = np.array([xmin, ymin - label_size[1]])
            else:
                text_origin = np.array([xmin, ymax])

            for i in range(thickness):
                draw.rectangle([xmin + i, ymin + i, xmax - i, ymax - i], outline=color)
            draw.rectangle(
                [tuple(text_origin), tuple(text_origin + label_size)], fill=color
            )
            draw.text(text_origin, thelabel, fill=(0, 0, 0), font=font)

    del draw

    return image



""" Using this File, we will be classifying Images using Predefined Models

and Scratch Models .

Author: Tushar Goel


Different Architectures:
    --> InceptionV3
    -->Xception
    --> VGG16
    --> VGG19
    --> Resnet50

"""
from .Dark_Neuron_CNN import Dark_Neuron
import  tensorflow as tf            # Powerful Framework for Deep Learning
import os                           # For Searching Folder within the system
from .models import Create_Model, Train_Model           # Script containing Different Models
from .Preprocessing_Image import Preprocess_Image      #Preprocessing Image Script
from .Prediction import Prediction
import matplotlib.pyplot as plt

 
class Classify_Images(Dark_Neuron):
    """
    This Class will have Following Properties:
        
    Attributes:
        --Working Directory
        -- Output Directory
        -- train --> Whether to train to predict
    Methods:
        --Preprocess_the_Images
        --Create_the_Model
        --Train_the_Model
        --Predict_from_the_Model
        --Visualize_the_Metric
        
    """
    def __init__(self,working_directory,output_directory):
        """
        In this function we will call Parent Function containing other Function
        
        and Define other variables.
        
        Arguments:
        ---------    
            working_directory --> Directory Containing Raw Data
            
            output_directory --> Output Sirectory to which Results will be posted
            
        Output:
        ------    
            None
        
        """
        Dark_Neuron.__init__(self,working_directory,output_directory)
        self.working_directory = working_directory
        self.output_directory = output_directory
        
    def load_model(self,user_model_name):
        user_model_path = os.path.join(self.working_directory,user_model_name)
        return tf.keras.models.load_model(user_model_path)

    """
    Defining Preprocess Function to Preprocess the Images with Different Flow Method
    
    """
    def Preprocess_the_Image(self,method,train,num_classes=2,batch_size=32,target_image_size=(224,224,3),model_name = None,user_model = None,image_path=None,grayscale=None,training_image_directory=None,validation_image_directory=None,dataframe=None,
                            test_image_directory=None,x_train=None,x_test=None,y_train=None,y_test=None,x_col=None,y_col = None,split=0.1,image_directory=None,input_tensor=None):
        """
        This function Will do image processing and return training Data Generator, Validation Data Generator
        
        and Test Data Generator on the Basis of Training Argument whether it is True Or False.
        
        Arguments:
            model_name --> Name for The Predefined Architecture
            num_classes --> Number of Classes
            batch_size --> Batch Size
            method --> Method by which Images will flow in the Function --> directory,dataframe,point
            training_image_directory --> Directory for method Directory
            validation_image_directory --> (Optional) For method Directory
            
            
        Outputs:
            It will Return the Data Generator for Train and Test
        
        """
        self.train = train
        self.target_image_size = target_image_size
        self.model_name = model_name
        self.num_classes = num_classes
        self.batch_size = batch_size
        self.method = method
        self.training_directory = training_image_directory
        self.validation_directory = validation_image_directory
        self.test_directory = test_image_directory
        self.dataframe = dataframe
        self.x_train = x_train
        self.x_test = x_test
        self.y_train = y_train
        self.y_test = y_test
        self.x_col_name = x_col
        self.y_col_name = y_col
        self.split = split
        self.image_directory = image_directory
        self.input_tensor = input_tensor
        self.image_path = image_path
        if user_model is not None:
            self.user_model = user_model
        else:
            self.user_model = None
        #Defining Variables for Preprocessing
        preprocessing = Preprocess_Image(model_name=self.model_name,user_model=self.user_model,target_image_size = self.target_image_size,num_classes = self.num_classes,batch_size=self.batch_size,training=self.train,method=self.method,working_directory = self.working_directory)
        
        #Getting results based on Different flow methods
        if self.method == 'directory':
            print('\n\t\t-----Getting Images From Directory------\n')
            if self.train:# From Preprocessing_Image.py File
                train_data,validation_data = preprocessing.Get_Images_from_Directory(self.training_directory,self.validation_directory,self.test_directory)
                print('\n\t\t-------Training Data Generated--------\n')
                return train_data,validation_data,(train_data.class_indices)
            else:
                test_data = preprocessing.Get_Images_from_Directory(self.training_directory,self.validation_directory,
                                                                    self.test_directory)
                print('\n\t\t-------Test Data Generated--------\n')
                return test_data
            
        elif self.method=='dataframe':
            print('\n\t\t-----Getting Images From DataFrame------\n')
            if self.train:
                train_data,validation_data = preprocessing.Get_Images_from_DataFrame(self.dataframe,self.x_col_name,self.split,self.y_col_name,self.image_directory)
                print('\n\t\t-----Training Data Generated------\n')
                
                return train_data,validation_data,(train_data.class_indices)
            
            else:
                test_data = preprocessing.Get_Images_from_DataFrame(self.dataframe,self.x_col_name,self.split,self.y_col_name,self.image_directory)
                print('\n\t\t-----Test Data Generated------\n')
                
                return test_data
        
        elif self.method=='point':
            print('\n\t\t-----Getting Images From Points------\n')
            if self.train:
                train_data,validation_data = preprocessing.Get_Data(self.x_train,self.y_train,self.x_test,self.y_test)
                print('\n\t\t------Training Data Generated-------\n')
                return train_data,validation_data
            else:
                test_data = preprocessing.Get_Data(self.x_train,self.y_train,self.x_test,self.y_test)
                print('\n\t\t---------Test Data Generated--------\n')
                return test_data
            
        elif self.method == 'image':
            if image_path is None:
                raise ValueError('Provide Image Path for Image Prediction or If it is containing in a directory having multiple ',
                                 'images, then set method = "directory"')
            print('\n\t\t----------Getting Image -------------\n')
            image = preprocessing.Get_Image(image_path = image_path,model_name = self.model_name,user_model=user_model,
                                            grayscale=grayscale)
            return image
       
        else:
            
            raise ValueError('Invalid Method Input --Must be from "directory","dataframe","point","image"')
            
            



    def Create_the_Model(self):
            
        """
        This Function will be used for Initialisation of Model according to Model name Given
        
        Arguments:
            None

            
        Returns:
            It will return the model for Training the model
            
        """
        print('\n\t\t--------------Model Creation Phase-----------\n')
    
        
        model_init = Create_Model(working_directory=self.working_directory,image_shape = self.target_image_size,train = self.train,input_tensor=self.input_tensor)
        
        # Defining Model based on Model name:
        if self.model_name in ['mobilenetv2','MobileNetV2','mobilenet_v2','MobileNet_V2']:
            
            # Checking whether Target Image size is within bounds for Predefined Architecture
            if self.target_image_size[0] <32 or self.target_image_size[1]<32:
                    Model_Target_Value_Checker()  #Check the Function Below which Raise the Value Error
                                 
            print('\n\t\t-------MobileNetV2 Model Initiated Successfully----------\n')
            return model_init.MobileNetV2()
            
        if self.model_name in ['inceptionv3','InceptionV3','inception_v3','Inception_V3']:
            
            # Checking whether Target Image size is within bounds for Predefined Architecture
            if self.target_image_size[0] <75 or self.target_image_size[1]<75:
                    Model_Target_Value_Checker()  #Check the Function Below which Raise the Value Error
            print('\n\t\t-------InceptiontV3 Model Initiated Successfully----------\n')                     
            return model_init.InceptionV3()
            
        if self.model_name in ['resnet50','ResNet50','Resnet50']:
            
            # Checking whether Target Image size is within bounds for Predefined Architecture
            if self.target_image_size[0] <32 or self.target_image_size[1]<32:
                    Model_Target_Value_Checker()  #Check the Function Below which Raise the Value Error
            print('\n\t\t-------Resnet50 Model Initiated Successfully----------\n')                     
            return model_init.ResNet50()
        
        if self.model_name in ['Xception','xception']:
            
            # Checking whether Target Image size is within bounds for Predefined Architecture
            if self.target_image_size[0] <71 or self.target_image_size[1]<71:
                    Model_Target_Value_Checker()  #Check the Function Below which Raise the Value Error
            print('\n\t\t-------Xception Model Initiated Successfully----------\n')                     
            return model_init.Xception()
        
        if self.model_name in ['VGG16','Vgg16','vgg16']:
            
            # Checking whether Target Image size is within bounds for Predefined Architecture
            if self.target_image_size[0] <32 or self.target_image_size[1]<32:
                    Model_Target_Value_Checker()  #Check the Function Below which Raise the Value Error
            print('\n\t\t-------VGG16 Model Initiated Successfully----------\n')                     
            return model_init.VGG16()
    
        if self.model_name in ['VGG19','Vgg19','vgg19']:
                    
                    # Checking whether Target Image size is within bounds for Predefined Architecture
                    if self.target_image_size[0] <32 or self.target_image_size[1]<32:
                            Model_Target_Value_Checker()  #Check the Function Below which Raise the Value Error
                    print('\n\t\t-------VGG19 Model Initiated Successfully----------\n')                     
                    return model_init.VGG19()            
            
            
            
            
        
    def Train_the_Model(self,model,rebuild=False,train_data_object=None,validation_data_object=None,test_data_object=None,epochs = 10,optimizer='adam',loss = 'binary_crossentropy',fine_tuning = False,layers = 20,metrics='accuracy',validation_steps=80,save_model = True,steps_per_epoch = 50,callbacks=None):
        """
        This function will call up the Initialised Model 
        
        """
        
        print('\n\t\t------------Model Training To be Start---------------')
        history,model = Train_Model(model=model,rebuild=rebuild,num_classes = self.num_classes,train_data_object=train_data_object,
                                    working_directory = self.working_directory,output_directory = self.output_directory,loss = loss,epochs=epochs,
                                    optimizer = optimizer,metrics = metrics,validation_data_object = validation_data_object,fine_tuning = fine_tuning,
                                    layers = layers,validation_steps=validation_steps,save_model=save_model,steps_per_epoch = steps_per_epoch,callbacks=callbacks)
        self.model_history = history
        return model
    
    def Visualize_the_Metrics(self):
        import matplotlib.pyplot as plt
        # Plot for Training Loss and Training Accuracy
        plt.plot(self.model_history.history['loss'],label='Training Loss')
        plt.plot(self.model_history.history['acc'],label = 'Training Accuracy')
        plt.title('Training Loss vs Training Accuracy')
        plt.legend()
        plt.show()
        #excetuted when validation set will be there
        try:
            # Training Loss vs Vaidation Loss 
            plt.plot(self.model_history.history['val_loss'],label='Test Loss')
            plt.plot(self.model_history.history['loss'],label = 'Training Loss')
            plt.title('Training Loss vs Validation Loss')
            plt.legend()
            plt.show()
            
            plt.plot(self.model_history.history['acc'],label='Training Accuracy')
            plt.plot(self.model_history.history['val_acc'],label = 'Validation Accuracy')
            plt.title('Training Accuracy vs Validation Accuracy')
            plt.legend()
            plt.show()
            
            plt.plot(self.model_history.history['val_loss'],label='Validation_Loss')
            plt.plot(self.model_history.history['val_acc'],label = 'Validation_Accuracy')
            plt.title('Validation Loss vs Validation Accuracy')
            plt.legend()
            plt.show()
        
        except:
            pass
    
    def Predict_from_the_Model(self,labels=None,generator=None,img = None,top = 5,model=None):
        """
        This Function will be used to predict the classes from Model
        
        Arguments:
            preprocessed_image --> preprocessed_image suitable for model
            model --> model get from trained part
            top --> Number of High Probabilities
            
        Return:
            Classes
        
        """
        self.generator = generator
        self.img = img
        prediction = Prediction(working_directory = self.working_directory,labels = labels,method = self.method,
                                model_name = self.model_name,user_model=self.user_model,
                                img = img,top=top,image_directory=self.image_path)
        
        if self.user_model is not None:
            model = self.user_model
        else:
            if model is None:
                raise ValueError('Provide Model, model argument should not be empty')
            model = model
        predicted_indices,predictions = prediction.prediction(method=self.method,model=model,img=img,data_generator=generator)
        
        label_score = prediction.label_class_provider(label=labels,predictions=predictions,predicted_indices=predicted_indices,
                                                      generator=generator,img=img)
        print('\n\t\t--------------Generating Predictions with Score----------------')
        
        self.label_score = label_score
        if len(label_score) == 0:
            print('\n\t\t----------No Predictions-----------')
            return label_score
        
        else:
            print(f'\n\t\t------------Found {len(label_score)} Predicitons-------' )
            return label_score
        
    def Visualize_the_Predictions(self,number=6):
        if number > len(self.label_score):
            number = len(self.label_score)
            
        if number ==0:
            print('No predictions to Show')
       
        
        else:
            if self.generator is not None:
                for label_score in self.label_score[:number]:
                    filepath = os.path.join(self.test_directory,label_score[0])
                    img = plt.imread(filepath)
                    plt.imshow(img)
                    plt.title(f'Predicted:{label_score[1].title()} ---- Score: {label_score[2]*100}')
                    plt.show()
            elif self.img is not None:
                for label_score in self.label_score[:number]:
                    filepath = label_score[0]
                    img = plt.imread(filepath)
                    plt.imshow(img)
                    plt.title(f'Predicted:{label_score[1].title()} ---- Score: {label_score[2]*100}')
                    plt.show()
            
            
def Model_Target_Value_Checker():
        raise ValueError('Try with Different Model.Get '
             'information on Keras Documentation\n'
             'The Lowest Dimensions allowed for Different Model are : \n'
             'Try to change in Preprocess Images Process \n'
             'MobileNetV2 --> (32,32) \n'
             'InceptionV3 --> (75,75)\n'
             'Resnet50 --> (32,32) \n'
             'Xception --> (71,71) \n'
             'VGG16 --> (32,32) \n'
             'VGG19 --> (32,32) \n'
             )





# vim: expandtab:ts=4:sw=4
import os
import errno
import argparse
import numpy as np
import cv2
import tensorflow as tf


def _run_in_batches(f, data_dict, out, batch_size):
    data_len = len(out)
    num_batches = int(data_len / batch_size)

    s, e = 0, 0
    for i in range(num_batches):
        s, e = i * batch_size, (i + 1) * batch_size
        batch_data_dict = {k: v[s:e] for k, v in data_dict.items()}
        out[s:e] = f(batch_data_dict)
    if e < len(out):
        batch_data_dict = {k: v[e:] for k, v in data_dict.items()}
        out[e:] = f(batch_data_dict)


def extract_image_patch(image, bbox, patch_shape):
    """Extract image patch from bounding box.

    Parameters
    ----------
    image : ndarray
        The full image.
    bbox : array_like
        The bounding box in format (x, y, width, height).
    patch_shape : Optional[array_like]
        This parameter can be used to enforce a desired patch shape
        (height, width). First, the `bbox` is adapted to the aspect ratio
        of the patch shape, then it is clipped at the image boundaries.
        If None, the shape is computed from :arg:`bbox`.

    Returns
    -------
    ndarray | NoneType
        An image patch showing the :arg:`bbox`, optionally reshaped to
        :arg:`patch_shape`.
        Returns None if the bounding box is empty or fully outside of the image
        boundaries.

    """
    bbox = np.array(bbox)
    if patch_shape is not None:
        # correct aspect ratio to patch shape
        target_aspect = float(patch_shape[1]) / patch_shape[0]
        new_width = target_aspect * bbox[3]
        bbox[0] -= (new_width - bbox[2]) / 2
        bbox[2] = new_width

    # convert to top left, bottom right
    bbox[2:] += bbox[:2]
    bbox = bbox.astype(np.int)

    # clip at image boundaries
    bbox[:2] = np.maximum(0, bbox[:2])
    bbox[2:] = np.minimum(np.asarray(image.shape[:2][::-1]) - 1, bbox[2:])
    if np.any(bbox[:2] >= bbox[2:]):
        return None
    sx, sy, ex, ey = bbox
    image = image[sy:ey, sx:ex]
    image = cv2.resize(image, tuple(patch_shape[::-1]))
    return image


class ImageEncoder(object):

    def __init__(self, checkpoint_filename, input_name="images",
                 output_name="features"):
        self.session = tf.Session()
        with tf.gfile.GFile(checkpoint_filename, "rb") as file_handle:
            graph_def = tf.GraphDef()
            graph_def.ParseFromString(file_handle.read())
        tf.import_graph_def(graph_def, name="net")
        self.input_var = tf.get_default_graph().get_tensor_by_name(
            "net/%s:0" % input_name)
        self.output_var = tf.get_default_graph().get_tensor_by_name(
            "net/%s:0" % output_name)

        assert len(self.output_var.get_shape()) == 2
        assert len(self.input_var.get_shape()) == 4
        self.feature_dim = self.output_var.get_shape().as_list()[-1]
        self.image_shape = self.input_var.get_shape().as_list()[1:]

    def __call__(self, data_x, batch_size=32):
        out = np.zeros((len(data_x), self.feature_dim), np.float32)
        _run_in_batches(
            lambda x: self.session.run(self.output_var, feed_dict=x),
            {self.input_var: data_x}, out, batch_size)
        return out


def create_box_encoder(model_filename, input_name="images",
                       output_name="features", batch_size=32):
    image_encoder = ImageEncoder(model_filename, input_name, output_name)
    image_shape = image_encoder.image_shape

    def encoder(image, boxes):
        image_patches = []
        for box in boxes:
            patch = extract_image_patch(image, box, image_shape[:2])
            if patch is None:
                print("WARNING: Failed to extract image patch: %s." % str(box))
                patch = np.random.uniform(
                    0., 255., image_shape).astype(np.uint8)
            image_patches.append(patch)
        image_patches = np.asarray(image_patches)
        return image_encoder(image_patches, batch_size)

    return encoder


def generate_detections(encoder, mot_dir, output_dir, detection_dir=None):
    """Generate detections with features.

    Parameters
    ----------
    encoder : Callable[image, ndarray] -> ndarray
        The encoder function takes as input a BGR color image and a matrix of
        bounding boxes in format `(x, y, w, h)` and returns a matrix of
        corresponding feature vectors.
    mot_dir : str
        Path to the MOTChallenge directory (can be either train or test).
    output_dir
        Path to the output directory. Will be created if it does not exist.
    detection_dir
        Path to custom detections. The directory structure should be the default
        MOTChallenge structure: `[sequence]/det/det.txt`. If None, uses the
        standard MOTChallenge detections.

    """
    if detection_dir is None:
        detection_dir = mot_dir
    try:
        os.makedirs(output_dir)
    except OSError as exception:
        if exception.errno == errno.EEXIST and os.path.isdir(output_dir):
            pass
        else:
            raise ValueError(
                "Failed to created output directory '%s'" % output_dir)

    for sequence in os.listdir(mot_dir):
        print("Processing %s" % sequence)
        sequence_dir = os.path.join(mot_dir, sequence)

        image_dir = os.path.join(sequence_dir, "img1")
        image_filenames = {
            int(os.path.splitext(f)[0]): os.path.join(image_dir, f)
            for f in os.listdir(image_dir)}

        detection_file = os.path.join(
            detection_dir, sequence, "det/det.txt")
        detections_in = np.loadtxt(detection_file, delimiter=',')
        detections_out = []

        frame_indices = detections_in[:, 0].astype(np.int)
        min_frame_idx = frame_indices.astype(np.int).min()
        max_frame_idx = frame_indices.astype(np.int).max()
        for frame_idx in range(min_frame_idx, max_frame_idx + 1):
            print("Frame %05d/%05d" % (frame_idx, max_frame_idx))
            mask = frame_indices == frame_idx
            rows = detections_in[mask]

            if frame_idx not in image_filenames:
                print("WARNING could not find image for frame %d" % frame_idx)
                continue
            bgr_image = cv2.imread(
                image_filenames[frame_idx], cv2.IMREAD_COLOR)
            features = encoder(bgr_image, rows[:, 2:6].copy())
            detections_out += [np.r_[(row, feature)] for row, feature
                               in zip(rows, features)]

        output_filename = os.path.join(output_dir, "%s.npy" % sequence)
        np.save(
            output_filename, np.asarray(detections_out), allow_pickle=False)


def parse_args():
    """Parse command line arguments.
    """
    parser = argparse.ArgumentParser(description="Re-ID feature extractor")
    parser.add_argument(
        "--model",
        default="resources/networks/mars-small128.pb",
        help="Path to freezed inference graph protobuf.")
    parser.add_argument(
        "--mot_dir", help="Path to MOTChallenge directory (train or test)",
        required=True)
    parser.add_argument(
        "--detection_dir", help="Path to custom detections. Defaults to "
        "standard MOT detections Directory structure should be the default "
        "MOTChallenge structure: [sequence]/det/det.txt", default=None)
    parser.add_argument(
        "--output_dir", help="Output directory. Will be created if it does not"
        " exist.", default="detections")
    return parser.parse_args()


def main():
    args = parse_args()
    encoder = create_box_encoder(args.model, batch_size=32)
    generate_detections(encoder, args.mot_dir, args.output_dir,
                        args.detection_dir)


if __name__ == "__main__":
    main()



"""
This Script contains Preprocessing Functions according to different Pre-Trained

architecture.

Different Architectures : Xception, MobileNetV2, InceptionV3, Resnet50

Author: Tushar Goel

"""
import tensorflow as tf
from tensorflow.keras.preprocessing import image
from tensorflow.keras.preprocessing.image import ImageDataGenerator
import numpy as np
from keras.utils.np_utils import to_categorical
from .Prediction import Prediction

class Preprocess_Image:
    """
    This Class will be Preprocessing Feature based on differnt Different Models and Differnt Methods for custom image 
    
    training
    
    Attributes:
    -----------
            model_name --> Differnt Architecture Models based On ImageNet 
            num_classes --> 2(Default) Number of Classes or Number of Objects to Process 
             
    Methods:
    --------
           Flow_from_Directory --> Taking Images from Data Directory
           Flow_from_DataFrame --> Taking Filename and Classes from DataFrame **Filename = Image_location
           Flow                --> Taking values x_train,y_train
           
    """
    
    def __init__(self,model_name=None,user_model = None,num_classes=2,batch_size=32,target_image_size=(224,224,3),training=False,method='directory',working_directory=None):
        
        self.user_model=user_model
        self.model_name=model_name
        self.method_name = method
        self.training = training
        if num_classes==2:
            self.class_mode='binary'                    # Creating Class Mode for Image Data Processing
        else:
            self.class_mode = 'categorical'
        self.batch_size = batch_size
        self.target_image_size = target_image_size
            
        print('\t\t----------------------------\n')
        print('\t\t Image Preprocessing Phase\n')
        print('\t\t----------------------------')
        
        self.pred = Prediction(method = self.method_name,user_model=self.user_model,model_name=self.model_name,working_directory = working_directory)
        
        
    def preprocess_architecture_function(self):
        """
        This Function will define preprocess function depends on the basis of the Model Architecture
        
        Arguments:
            None
        
        Output:
            Preprocess_function
        
        """
        #Based on Different Architecture of Models:
        if self.user_model is not None:
            pass
        else:
            if self.model_name is None:
                raise ValueError('Provide Model name or Provide user model')
            if self.model_name in ['mobilenetv2','MobileNetV2','mobilenet_v2','MobileNet_V2']:
                preprocess_function = tf.keras.applications.mobilenet_v2.preprocess_input
                return preprocess_function
            
            elif self.model_name in ['resnet50','ResNet50','Resnet50']:
                preprocess_function = tf.keras.applications.resnet50.preprocess_input
                return preprocess_function
            
            elif self.model_name in ['inceptionv3','InceptionV3','inception_v3','Inception_V3']:
                preprocess_function = tf.keras.applications.inception_v3.preprocess_input
                return preprocess_function
            
            elif self.model_name in ['VGG16','Vgg16','vgg16']:
                preprocess_function = tf.keras.applications.vgg16.preprocess_input
                return preprocess_function
            
            elif self.model_name in ['Xception','xception']:
                preprocess_function = tf.keras.applications.xception.preprocess_input
                return preprocess_function
            
            elif self.model_name in ['VGG19','Vgg19','vgg19']:
                preprocess_function = tf.keras.applications.vgg19.preprocess_input
                return preprocess_function
            else:
                raise ValueError('Invalid Model Name')
        
    def Get_Images_from_Directory(self,training_images_directory=None,validation_images_directory=None,test_image_directory=None):
        
        """
        This Function will Take images from image Directory and convert them to desired
        
        shape and Configurations.
        
        Arguments:
            
            training_images_directory --> Directory Containing Images from Training Folder for Preprocessing
            
            Validation_images_directory --> None (Defalult)--No Validation will be occur.
            
            batch_size --> Batch Size for Data Generator
            
            target_image_size --> Contains Target Size for conversion of Model including 3 channel
            
        Output:
            
            Training_Data_Generator,Validation_data_Generator
            
        """
        
        target_image_size = self.target_image_size
        if len(target_image_size)!=3:
            raise ValueError('Required 3 Arguments --- {} are given'.format(len(target_image_size)))
            
        #Defining Preprocessing Function
        if self.user_model is not None:
            preprocessing_function = None
        else:
            preprocessing_function = self.preprocess_architecture_function()
        # Data Generator Function for preprocessing Function
        data_generator = ImageDataGenerator(preprocessing_function = preprocessing_function,
                                                rescale=1./255,
                                                shear_range=0.2,
                                                zoom_range=0.2,
                                                horizontal_flip=True)
        
        if self.training:
            if training_images_directory is None:
                raise ValueError('For Training Image Directory is Must, or put training mode to False and use test_image_directory')
            # Train Data Generator for Training
            
            if self.user_model is not None:
                self.target_image_size = self.user_model.input_shape[1:3]
            train_data_generator = data_generator.flow_from_directory(directory = training_images_directory,
                                                                      target_size = (self.target_image_size[0],self.target_image_size[1]),
                                                                      batch_size = self.batch_size,
                                                                      class_mode = self.class_mode)
            #Checking Validation Data Directory , If exists return train and validation data generator
            if validation_images_directory is not None:
                validation_data_generator = data_generator.flow_from_directory(directory = validation_images_directory,
                                                                               target_size = (target_image_size[0],target_image_size[1]),
                                                                               class_mode = self.class_mode)
                return train_data_generator,validation_data_generator
            
            else:
                return train_data_generator,None
        else:
            if test_image_directory is None:
                raise ValueError('Test Image Directory Required for Prediction')
            # Data Generator Function for preprocessing Function
            target_size = self.pred.target_shape_generator()
            data_generator = ImageDataGenerator(
                                                rescale = 1./255)
            test_data_generator = data_generator.flow_from_directory(directory = test_image_directory,
                                                                     target_size = target_size,
                                                                     shuffle=False,
                                                                     batch_size = 1
                                                                     )
            return test_data_generator
            
    def Get_Images_from_DataFrame(self,dataframe,x_column_name,split=0.1,y_column_name=None,image_directory=None):
        """
        This Function will take Filename from Dataframe and Preprocess it from that loacation 
        
        and store corresponding Class value
        
        Arguments:
            
            Dataframe --> This DataFrame Contains Filename of Images and Class Variables
            
            x_column_name --> This Will contain the column name of Filename containg Column
            
            y_column_name --> This will contain the column name of Classes Containing column
                                   
        Outputs:
            
            Training data Generator, Validation Data Generator
            
        """
        
        #Defining Preprocess Function
        preprocessing_function = self.preprocess_architecture_function()
        
        # Data Generator Function for preprocessing Function
        data_generator = ImageDataGenerator(preprocessing_function = preprocessing_function,
                                            rescale = 1./255)
        
        #Splitting Functions for training and Predictions :
        if self.training:
            dataframe_length = dataframe.shape[0]
            
            #Splitting dataframe for Validation and Training Data 
            training_length = round((1-split)*dataframe_length)
            training_dataframe = dataframe[:training_length]
            validation_dataframe = dataframe[training_length:]
            validation_dataframe.columns = dataframe.columns.values
            
            if y_column_name is None:
                raise ValueError(' For Training Y Columns is required')
            # Train Data Generator for Training
            if self.user_model is not None:
                self.target_image_size = self.user_model.input_shape[1:3]
            train_data_generator = data_generator.flow_from_dataframe(dataframe=training_dataframe,
                                                                      directory = image_directory,
                                                                      x_col = x_column_name,
                                                                      y_col = y_column_name,
                                                                      target_size = (self.target_image_size[0],self.target_image_size[1]),
                                                                      class_mode = self.class_mode,
                                                                      
                                                                      batch_size = self.batch_size)
            try:
                validation_data_generator = data_generator.flow_from_dataframe(dataframe=validation_dataframe,
                                                                          directory = image_directory,
                                                                          x_col = x_column_name,
                                                                          y_col = y_column_name,
                                                                          target_size = (self.target_image_size[0],self.target_image_size[1]),
                                                                          class_mode = self.class_mode,
                                                                
                                                                          batch_size = self.batch_size)
                print('-----Validation and Train Data Generated----------')
                return train_data_generator,validation_data_generator
            except:
                return train_data_generator,None
        
        else:
            # Test Data Generator 
            target_size = self.pred.target_shape_generator()
            test_data_generator = data_generator.flow_from_dataframe(dataframe=dataframe,
                                                                      directory = image_directory,
                                                                      x_col = x_column_name,
                                                                      y_col = y_column_name,
                                                                      target_size = target_size,
                                                                      shuffle=False,
                                                                      batch_size = 1
                                                                     )
            return test_data_generator
        
    def Get_Data(self,x_train=None,y_train=None,x_test=None,y_test=None):
        """
        This Function will take value from arrays containing Data and Labels
        
        Arguments:
            x_training --> Containg Image Data in form of arrays
            y --> Containg Labels for Image Data
            
        Output:
            Train Data generator , Validation Data Generator
            
        """
        #Defining Preprocess Function
        preprocessing_function = self.preprocess_architecture_function()
        
        # Data Generator Function for preprocessing Function
        data_generator = ImageDataGenerator(preprocessing_function,
                                            rescale = 1./255)
        if len(x_train.shape)==3:
            x_train = np.expand_dims(x_train,axis=-1)
            x_test = np.expand_dims(x_test,axis=-1)
        
        y_train = to_categorical(y_train)
        y_test = to_categorical(y_test)
        if self.training:
             # Train Data Generator for x_train,y_train
            if y_train is None:
                raise ValueError('For Training, Classes are required')
            train_data_generator = data_generator.flow(x = x_train,
                                                       y = y_train,
                                                
                                                        batch_size = self.batch_size)
            validation_data_generator = data_generator.flow(x = x_test,
                                                       y = y_test,
                                                    
                                                       batch_size = self.batch_size)
            return train_data_generator,validation_data_generator
        
        else:
            test_data_generator = data_generator.flow(x = x_test,
                                                       y = y_test
                                                       )
            
            return test_data_generator
        
    
    def Get_Image(self,image_path,model_name = None,user_model= None,grayscale = False):
        """
        This Function will be for Image Processing for Single Images not on any Directory
        
        Arguments:
            model_name --> Model Name for PreDefined Architecture: None(Default)
            user_model --> Path to user own Model if they have Predefined Models : None(Default)
            
            target_image_size --> This will be the target image size user model or Predefined Architecture Model
            image_path --> Image Path which is to be Predicted
    
        Returns:
            
            This Function will Return the Preprocessed Image
            
        """
        target_shape = self.pred.target_shape_generator()
        
        if user_model is not None:
            
            img = image.load_img(image_path,grayscale=grayscale,target_size=target_shape)
            img = image.img_to_array(img)
            img = np.expand_dims(img,axis=0)
            
            return img
        
        else:
             # Recalling Function from Above for different models
            
            img = image.load_img(image_path,grayscale=grayscale,target_size=target_shape)
            img = image.img_to_array(img)
            img = np.expand_dims(img,axis=0)
            
            return img

            

        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
    
    



from PIL import Image
import numpy as np
import cv2
from .yolo import YOLO
import os 


def Web_Cam_Detection(working_directory,videopath=0,model_name=None,classes = [],score = 0.6,iou = 0.5,gpu_num = 1):
    
    anchors_path = os.path.join(os.path.dirname(__file__),'yolo4_anchors.txt')
    if model_name is None:
        model_path = os.path.join(working_directory,'yolov4.h5')
        classes_path = os.path.join(os.path.dirname(__file__),'coco_classes.txt')
    else:
        model_path = os.path.join(working_directory,model_name)
        classes_path = os.path.join(working_directory,'data_classes.txt')
    
    yolo = YOLO(
        **{
            "model_path": model_path,
            "anchors_path": anchors_path,
            "classes_path": classes_path,
            "score": score,
            "gpu_num": gpu_num,
            "model_image_size": (608,608),
        }
        )
    cap = cv2.VideoCapture(videopath)
    out = cv2.VideoWriter(os.path.join(working_directory,'output_web.avi'),cv2.VideoWriter_fourcc(*"MJPG"), 5, (int(cap.get(3)), int(cap.get(4))))    
    
    while True:
        ret,frame = cap.read()
        
        img = Image.fromarray(frame)
        pred,image = yolo.detect_image(img,classes = classes,score=score)
        
        image = np.array(image)
        out.write(image)
        cv2.namedWindow('image',cv2.WINDOW_NORMAL)
        cv2.resizeWindow('image',(600,600))
        cv2.imshow('image',image)
        
        if cv2.waitKey(27) & 0xFF == ord('q'):
            break
        
    cap.release()
    out.release()
    cv2.destroyAllWindows()
        


"""
This Script will be used to download Darknet YOLOv3 Weights and Convert them into

Keras Model yolo.h5 

Author: Tushar Goel

"""
import os
import colorsys
import sys
import numpy as np
from keras import backend as K
from keras.models import load_model
from keras.layers import Input
from keras.utils import multi_gpu_model
from .yolo4_model import yolo_eval, yolo4_body
from .yolo4_utils import letterbox_image
import wget
from PIL import Image, ImageFont, ImageDraw
from timeit import default_timer as timer
import matplotlib.pyplot as plt

from operator import itemgetter

class Yolo4_weights(object):
    
    def __init__(self, score, iou, anchors_path, classes_path, model_path, weights_path, gpu_num=1):
        self.score = score
        self.iou = iou
        self.anchors_path = anchors_path
        self.classes_path = classes_path
        self.weights_path = weights_path
        self.model_path = model_path
        self.gpu_num = gpu_num
        self.load_yolo()
    
    
    
    def get_class(self):
        
        classes_path = os.path.expanduser(self.classes_path)
        with open(classes_path) as f:
            class_names = f.readlines()
        class_names = [c.strip() for c in class_names]
        return class_names

    def get_anchors(self):
        
        anchors_path = os.path.expanduser(self.anchors_path)
        with open(anchors_path) as f:
            anchors = f.readline()
        anchors = [float(x) for x in anchors.split(',')]
        return np.array(anchors).reshape(-1, 2)

    def load_yolo(self):
        
        model_path = os.path.expanduser(self.model_path)
        assert model_path.endswith('.h5'), 'Keras model or weights must be a .h5 file.'

        self.class_names = self.get_class()
        self.anchors = self.get_anchors()

        num_anchors = len(self.anchors)
        num_classes = len(self.class_names)

        # Generate colors for drawing bounding boxes.
        hsv_tuples = [(x / len(self.class_names), 1., 1.)
                      for x in range(len(self.class_names))]
        self.colors = list(map(lambda x: colorsys.hsv_to_rgb(*x), hsv_tuples))
        self.colors = list(
            map(lambda x: (int(x[0] * 255), int(x[1] * 255), int(x[2] * 255)),
                self.colors))

        self.sess = K.get_session()

        # Load model, or construct model and load weights.
        self.yolo4_model = yolo4_body(Input(shape=(608, 608, 3)), num_anchors//3, num_classes)

        # Read and convert darknet weight
        print('Loading weights.')
        weights_file = open(self.weights_path, 'rb')
        major, minor, revision = np.ndarray(
            shape=(3, ), dtype='int32', buffer=weights_file.read(12))
        if (major*10+minor)>=2 and major<1000 and minor<1000:
            seen = np.ndarray(shape=(1,), dtype='int64', buffer=weights_file.read(8))
        else:
            seen = np.ndarray(shape=(1,), dtype='int32', buffer=weights_file.read(4))
        print('Weights Header: ', major, minor, revision, seen)

        convs_to_load = []
        bns_to_load = []
        for i in range(len(self.yolo4_model.layers)):
            layer_name = self.yolo4_model.layers[i].name
            if layer_name.startswith('conv2d_'):
                convs_to_load.append((int(layer_name[7:]), i))
            if layer_name.startswith('batch_normalization_'):
                bns_to_load.append((int(layer_name[20:]), i))

        convs_sorted = sorted(convs_to_load, key=itemgetter(0))
        bns_sorted = sorted(bns_to_load, key=itemgetter(0))

        bn_index = 0
        for i in range(len(convs_sorted)):
            print('Converting ', i)
            if i == 93 or i == 101 or i == 109:
                #no bn, with bias
                weights_shape = self.yolo4_model.layers[convs_sorted[i][1]].get_weights()[0].shape
                bias_shape = self.yolo4_model.layers[convs_sorted[i][1]].get_weights()[0].shape[3]
                filters = bias_shape
                size = weights_shape[0]
                darknet_w_shape = (filters, weights_shape[2], size, size)
                weights_size = np.product(weights_shape)

                conv_bias = np.ndarray(
                    shape=(filters, ),
                    dtype='float32',
                    buffer=weights_file.read(filters * 4))
                conv_weights = np.ndarray(
                    shape=darknet_w_shape,
                    dtype='float32',
                    buffer=weights_file.read(weights_size * 4))
                conv_weights = np.transpose(conv_weights, [2, 3, 1, 0])
                self.yolo4_model.layers[convs_sorted[i][1]].set_weights([conv_weights, conv_bias])
            else:
                #with bn, no bias
                weights_shape = self.yolo4_model.layers[convs_sorted[i][1]].get_weights()[0].shape
                size = weights_shape[0]
                bn_shape = self.yolo4_model.layers[bns_sorted[bn_index][1]].get_weights()[0].shape
                filters = bn_shape[0]
                darknet_w_shape = (filters, weights_shape[2], size, size)
                weights_size = np.product(weights_shape)

                conv_bias = np.ndarray(
                    shape=(filters, ),
                    dtype='float32',
                    buffer=weights_file.read(filters * 4))
                bn_weights = np.ndarray(
                    shape=(3, filters),
                    dtype='float32',
                    buffer=weights_file.read(filters * 12))

                bn_weight_list = [
                    bn_weights[0],  # scale gamma
                    conv_bias,  # shift beta
                    bn_weights[1],  # running mean
                    bn_weights[2]  # running var
                ]
                self.yolo4_model.layers[bns_sorted[bn_index][1]].set_weights(bn_weight_list)

                conv_weights = np.ndarray(
                    shape=darknet_w_shape,
                    dtype='float32',
                    buffer=weights_file.read(weights_size * 4))
                conv_weights = np.transpose(conv_weights, [2, 3, 1, 0])
                self.yolo4_model.layers[convs_sorted[i][1]].set_weights([conv_weights])

                bn_index += 1

        weights_file.close()

        self.yolo4_model.save(self.model_path)


        if self.gpu_num>=2:
            self.yolo4_model = multi_gpu_model(self.yolo4_model, gpus=self.gpu_num)

        self.input_image_shape = K.placeholder(shape=(2, ))
        self.boxes, self.scores, self.classes = yolo_eval(self.yolo4_model.output, self.anchors,
                len(self.class_names), self.input_image_shape,
                score_threshold=self.score)

    

    def close_session(self):
        self.sess.close()


def bar_progress(current, total, width=80):
  progress_message = "Downloading: %d%% [%d / %d] bytes" % (current / total * 100, current, total)
  # Don't use print() as it will print in new line every time.
  sys.stdout.write("\r" + progress_message)
  sys.stdout.flush()

def Download_weights(working_directory):
        """
        This Function will be used to download yolov3 weights from Darknet.
        
        """
    
        weights_path = os.path.join(working_directory,"yolov4.weights")
        
        if not os.path.exists(weights_path):
            
            print('Downloading YoloV4 Weights...if you have downloaded weights,'
                  ' then stop the process and put it in working directory with name'
                  ' yolov3.weights')
            url = "https://github.com/AlexeyAB/darknet/releases/download/darknet_yolo_v3_optimal/yolov4.weights"
            wget.download(url,weights_path,bar=bar_progress)
            
            print('Weights Downloaded')
        
        print('Weights Loaded......Converting them to Keras Model')
  



from setuptools import setup

def readme():
    with open('README.md') as f:
        README = f.read()
    return README

setup(
  name = 'DarkNeurons',         
  packages = ['DarkNeurons'],  


  version = '1.3.4',    
  long_description=readme(),
  long_description_content_type="text/markdown",     


  license='MIT',       
  description = 'A Library for Easy Implementation of Deep learning Techniques',   
  author = 'Tushar Goel',
  author_email = 'tgoel219@gmail.com',     
  url = 'https://github.com/Tushar-ml/DarkNeuron',   


  download_url = 'https://github.com/Tushar-ml/DarkNeuron/archive/v1.3.4.tar.gz',    
    
  keywords = ['Keras','Object_Detection','CNN'],   # Keywords that define your package best
  install_requires=[            # I get to this in a second
          'wget','tensorflow >= 1.15.0','numpy >= 1.18.4','scipy >= 1.4.1','matplotlib >= 3.2.1',
          'keras >= 2.1.4', 'pandas>=0.23.0','PyAutoGUI>=0.9.48','opencv-python>=4.2.0.32','Pillow>=5.3.0'
      ],
  classifiers=[
    'License :: OSI Approved :: MIT License',   # Again, pick a license
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3'
  ],
)




"""
Run a YOLO_v3 style detection model on test images.
"""

import colorsys
import os
import random
from timeit import time
from timeit import default_timer as timer  ### to calculate FPS

import numpy as np
from keras import backend as K
from keras.models import load_model
from PIL import Image, ImageFont, ImageDraw

from .yolo4_model import yolo_eval
from .yolo4_utils import letterbox_image

class YOLO_Tracker(object):
    def __init__(self,working_directory,model_name=None,score=0.6,classes = []):
        if model_name is None:
            self.model_path = os.path.join(working_directory,'yolov4.h5')
            self.classes_path = os.path.join(os.path.dirname(__file__),'model_data/coco_classes.txt')
        
        else:
            self.model_path = os.path.join(working_directory,model_name)
            self.classes_path = os.path.join(working_directory,'data_classes.txt')
        self.anchors_path = os.path.join(os.path.dirname(__file__),'model_data/yolo4_anchors.txt')
        
        self.score = score
        self.iou = 0.5
        self.class_names = self._get_class()
        self.anchors = self._get_anchors()
        self.sess = K.get_session()
        self.model_image_size = (416, 416) # fixed size or (None, None)
        self.is_fixed_size = self.model_image_size != (None, None)
        self.boxes, self.scores, self.classes = self.generate()
        self.selected_classes = classes
    def _get_class(self):
        classes_path = os.path.expanduser(self.classes_path)
        with open(classes_path) as f:
            class_names = f.readlines()
        class_names = [c.strip() for c in class_names]
        return class_names

    def _get_anchors(self):
        anchors_path = os.path.expanduser(self.anchors_path)
        with open(anchors_path) as f:
            anchors = f.readline()
            anchors = [float(x) for x in anchors.split(',')]
            anchors = np.array(anchors).reshape(-1, 2)
        return anchors

    def generate(self):
        model_path = os.path.expanduser(self.model_path)
        assert model_path.endswith('.h5'), 'Keras model must be a .h5 file.'

        self.yolo_model = load_model(model_path, compile=False)
        print('{} model, anchors, and classes loaded.'.format(model_path))

        # Generate colors for drawing bounding boxes.
        hsv_tuples = [(x / len(self.class_names), 1., 1.)
                      for x in range(len(self.class_names))]
        self.colors = list(map(lambda x: colorsys.hsv_to_rgb(*x), hsv_tuples))
        self.colors = list(
            map(lambda x: (int(x[0] * 255), int(x[1] * 255), int(x[2] * 255)),
                self.colors))
        random.seed(10101)  # Fixed seed for consistent colors across runs.
        random.shuffle(self.colors)  # Shuffle colors to decorrelate adjacent classes.
        random.seed(None)  # Reset seed to default.

        # Generate output tensor targets for filtered bounding boxes.
        self.input_image_shape = K.placeholder(shape=(2, ))
        boxes, scores, classes = yolo_eval(self.yolo_model.output, self.anchors,
                len(self.class_names), self.input_image_shape,
                score_threshold=self.score, iou_threshold=self.iou)
        return boxes, scores, classes

    def detect_image(self, image):
        if self.is_fixed_size:
            assert self.model_image_size[0]%32 == 0, 'Multiples of 32 required'
            assert self.model_image_size[1]%32 == 0, 'Multiples of 32 required'
            boxed_image = letterbox_image(image, tuple(reversed(self.model_image_size)))
        else:
            new_image_size = (image.width - (image.width % 32),
                              image.height - (image.height % 32))
            boxed_image = letterbox_image(image, new_image_size)
        image_data = np.array(boxed_image, dtype='float32')

        #print(image_data.shape)
        image_data /= 255.
        image_data = np.expand_dims(image_data, 0)  # Add batch dimension.
        
        out_boxes, out_scores, out_classes = self.sess.run(
            [self.boxes, self.scores, self.classes],
            feed_dict={
                self.yolo_model.input: image_data,
                self.input_image_shape: [image.size[1], image.size[0]],
                K.learning_phase(): 0
            })
        return_boxs = []
        return_scores = []
        for i, c in reversed(list(enumerate(out_classes))):
            if (len(self.selected_classes) != 0):
                if self.class_names[c] in self.selected_classes:
                    predicted_class = self.class_names[c]
                else:
                    continue
            else:
                predicted_class = self.class_names[c]
            box = out_boxes[i]
            if out_scores[i]>self.score:
                score = out_scores[i]
                
            else:
                continue
            box = out_boxes[i]
            score = out_scores[i]
            x = int(box[1])  
            y = int(box[0])  
            w = int(box[3]-box[1])
            h = int(box[2]-box[0])
            if x < 0 :
                w = w + x
                x = 0
            if y < 0 :
                h = h + y
                y = 0 
            return_boxs.append([x,y,w,h])
            return_scores.append(score)

        return return_boxs, return_scores

    def close_session(self):
        self.sess.close()


"""
This Script is For Prediction . In this, Prediction will be filter out 

from Bunch of Raw data

Author: Tushar Goel

"""
import tensorflow as tf
from .models import Create_Model
from tensorflow.keras.preprocessing import image
import numpy as np


class Prediction:
    """
    In this Object, we will predict the Untested Objects and Provide the Filtered Predictions
    
    Attributes:
        method: Whether they are providing Dircetory,dataframe,or Single Images
        model_name: to Predict directly from the imagenet models
        
    """
    def __init__(self,method,working_directory,labels=None,model_name=None,user_model=None,img=None,directory = None,top=None,image_directory=None):
        
        self.labels = labels
        self.model_name = model_name
        self.user_model = user_model
        self.working_directory = working_directory
        self.top = top
        self.image_directory=image_directory
        print('\n\t\t-------------Prediction Phase--------------')
        
    def target_shape_generator(self):
        
        model_init = Create_Model(working_directory = self.working_directory)
        if self.user_model is not None:
            target_shape = self.user_model.input_shape[1:3]
            print('\n\t\t----------Model for Prediction Ready----------')
            return target_shape
        
        
        else:
            if self.model_name is None:
                raise ValueError('Provide atleast user model or model_name')
            # Defining Model based on Model name:
            if self.model_name in ['mobilenetv2','MobileNetV2','mobilenet_v2','MobileNet_V2']:
    
                print('\n\t\t-------MobilenetV2 For Prediction Ready----------\n')
                
                return model_init.MobileNetV2().input_shape[1:3]
                
            if self.model_name in ['inceptionv3','InceptionV3','inception_v3','Inception_V3']:
                
                
                print('\n\t\t-------InceptiontV3 for Prediction Ready----------\n')
                model = model_init.InceptionV3()                     
                return model.input_shape[1:3]
                
            if self.model_name in ['resnet50','ResNet50','Resnet50']:
    
                print('\n\t\t-------Resnet50 Model for Prediction Ready----------\n')                     
                return model_init.ResNet50().input_shape[1:3]
            
            if self.model_name in ['Xception','xception']:
    
                print('\n\t\t-------Xception Model for Prediction Ready----------\n')                     
                return model_init.Xception().input_shape[1:3]
            
            if self.model_name in ['VGG16','Vgg16','vgg16']:
                
    
                print('\n\t\t-------VGG16 Model for Prediction Ready----------\n')                     
                return model_init.VGG16().input_shape[1:3]
        
            if self.model_name in ['VGG19','Vgg19','vgg19']:
                        
            
                print('\n\t\t-------VGG19 Model for Prediction Ready----------\n')                     
                return model_init.VGG19()
        
        
        
    def prediction(self,method,model,img=None,data_generator=None):
        
        if self.model_name is None:
            #model = tf.keras.models.load_model(model)        
            if method in ['directory','dataframe','point']:
                if data_generator is None:
                    raise ValueError('Provide Image Generators in generator argument')
                prediction = model.predict_generator(data_generator,steps=len(data_generator),verbose=1)
                predicted_indices = np.argmax(prediction,axis=1)
                return predicted_indices,prediction
            
            elif method in ['image']:
                prediction = model.predict(img)
                predicted_indices = np.argmax(prediction,axis=1)
                return predicted_indices,prediction
        
        else:
            if self.model_name in ['mobilenetv2','MobileNetV2','mobilenet_v2','MobileNet_V2']:
    
                print('\n\t\t-------Getting Label For MobileNetV2----------\n')
                
                decode_predictions = tf.keras.applications.mobilenet_v2
                
            if self.model_name in ['inceptionv3','InceptionV3','inception_v3','Inception_V3']:
                
                
               print('\n\t\t-------Getting Label For InceptionV3----------\n')
                
               decode_predictions = tf.keras.applications.inception_v3
                
            if self.model_name in ['resnet50','ResNet50','Resnet50']:
    
                print('\n\t\t-------Getting Label For ResNet50----------\n')
                
                decode_predictions = tf.keras.applications.resnet50
            
            if self.model_name in ['Xception','xception']:
    
                print('\n\t\t-------Getting Label For Xception----------\n')
                
                decode_predictions = tf.keras.applications.xception
            
            if self.model_name in ['VGG16','Vgg16','vgg16']:
                
    
                print('\n\t\t-------Getting Label For VGG16----------\n')
                
                decode_predictions = tf.keras.applications.vgg16
        
            if self.model_name in ['VGG19','Vgg19','vgg19']:
                        
            
                print('\n\t\t-------Getting Label For MobileNetV2----------\n')
                
                decode_predictions = tf.keras.applications.vgg19
            if method in ['directory','dataframe','point']:
                if data_generator is None:
                    raise ValueError('Provide Image Generators in generator argument')
                prediction = model.predict_generator(data_generator,steps=len(data_generator),verbose=1)
                predicted_indices = decode_predictions.decode_predictions(prediction,top=self.top)
                return predicted_indices,prediction
            
            elif method in ['image']:
                prediction = model.predict(img)
                predicted_indices = decode_predictions.decode_predictions(prediction,top=self.top)
                return predicted_indices,prediction   
            
    def label_class_provider(self,predictions,predicted_indices,label=None,generator=None,img=None):
        
        if type(label) is list:
            class_names = sorted(label) # Sorting them
            label = dict(zip(class_names, range(len(class_names))))
        print("Following are the Predictions made\n")
        print(predicted_indices)
        if self.model_name is None:
            if generator is not None:
            
                labels_score = []
                for i in range(len(predicted_indices)):#Generating Predictions Labels Scores with their Labels
                    for key,value in label.items():
                        if value == predicted_indices[i]:
                            labels_score.append([generator.filenames[i],key,predictions[i][predicted_indices[i]]])
                return labels_score
            
            else:
                if img is None:
                    raise ValueError('Provide Data_Generator or File name ')
                labels_score = []
                
                for key,value in label.items():
                    
                    if value == predicted_indices:
                        labels_score.append([self.image_directory,key,max(predictions)[0]])
                return labels_score
        else:
             if generator is not None:
            
                labels_score = []
                for i in range(len(predicted_indices)):#Generating Predictions Labels Scores with their Labels
  
                            labels_score.append([generator.filenames[i],predicted_indices[i][0][1],predicted_indices[i][0][2]])
                return labels_score
            
             else:
                if img is None:
                    raise ValueError('Provide Data_Generator or File name ')
                labels_score = []
     
                labels_score.append([self.image_directory,predicted_indices[0][0][1],predicted_indices[0][0][2]])
                return labels_score
        
        
        
            
        


"""
This Script is used for Detection purposes whether it's video or images

and also be used for Real-Time Detection

Author: Tushar Goel

"""

import os
import sys
from .yolo import YOLO,detect_video
from PIL import Image
from timeit import default_timer as timer
from .utils import load_extractor_model, load_features, parse_input, detect_object
import pandas as pd
import numpy as np
from .Yolo_Format import GetFileList
import random


def Detector(working_directory,test_folder_name,classes = [],model_name=None,score=0.5,gpu_num = 1):
    """
    This Function will be used for Detection of Objects in Video Or Images
    
    Arguments:
        working_directory --> Working Directory where weights and Test Images are Kept.
        Test_Folder_name -->  Name of the Test Folder name.
        model_name --> Name of the Model(None--> using yolov4.h5)
        
    Return:
        Detections
        
    """
    image_test_folder = os.path.join(working_directory,test_folder_name)

    if model_name is None:
        model_weights = os.path.join(working_directory,'yolov4.h5')
        model_classes = os.path.join(os.path.dirname(__file__),'coco_classes.txt')
    else:
        model_weights = os.path.join(working_directory,model_name)
        model_classes = os.path.join(working_directory,'data_classes.txt')
    
    anchors_path = os.path.join(os.path.dirname(__file__),'yolo4_anchors.txt')
    detection_results_file = os.path.join(working_directory,'Detections_results.csv')
    postfix = 'Detection'
    save_img = True
    
    input_paths = GetFileList(dirName = image_test_folder)
    img_endings = (".jpg", ".jpeg", ".png")
    vid_endings = (".mp4", ".mpeg", ".mpg", ".avi",".mkv")

    input_image_paths = []
    input_video_paths = []
    for item in input_paths:
        if item.endswith(img_endings):
            input_image_paths.append(item)
        elif item.endswith(vid_endings):
            input_video_paths.append(item)

    output_path = os.path.join(working_directory,'Test_results')
    if not os.path.exists(output_path):
        os.makedirs(output_path)
        
    # define YOLO detector
    yolo = YOLO(
        **{
            "model_path": model_weights,
            "anchors_path": anchors_path,
            "classes_path": model_classes,
            "score": score,
            "gpu_num": gpu_num,
            "model_image_size": (608,608),
        }
    )
    
    # Make a dataframe for the prediction outputs
    out_df = pd.DataFrame(
        columns=[
            "image",
            "image_path",
            "xmin",
            "ymin",
            "xmax",
            "ymax",
            "label",
            "confidence",
            "x_size",
            "y_size",
        ]
    )
    
    # labels to draw on images
    class_file = open(model_classes, "r")
    input_labels = [line.rstrip("\n") for line in class_file.readlines()]
    print("Found {} input labels: {} ...".format(len(input_labels), input_labels))

    if input_image_paths:
        print(
            "Found {} input images: {} ...".format(
                len(input_image_paths),
                [os.path.basename(f) for f in input_image_paths[:5]],
            )
        )
        start = timer()
        text_out = ""
        
        # This is for images
        for i, img_path in enumerate(input_image_paths):
            print(img_path)
            prediction, image = detect_object(
                yolo,
                img_path,
                save_img=save_img,
                save_img_path=output_path,
                postfix=postfix,
            )
            y_size, x_size, _ = np.array(image).shape
            for single_prediction in prediction:
                out_df = out_df.append(
                    pd.DataFrame(
                        [
                            [
                                os.path.basename(img_path.rstrip("\n")),
                                img_path.rstrip("\n"),
                            ]
                            + single_prediction
                            + [x_size, y_size]
                        ],
                        columns=[
                            "image",
                            "image_path",
                            "xmin",
                            "ymin",
                            "xmax",
                            "ymax",
                            "label",
                            "confidence",
                            "x_size",
                            "y_size",
                        ],
                    )
                )
        end = timer()
        print(
            "Processed {} images in {:.1f}sec - {:.1f}FPS".format(
                len(input_image_paths),
                end - start,
                len(input_image_paths) / (end - start),
            )
        )
        out_df.to_csv(detection_results_file, index=False)
        
    # This is for videos
    if input_video_paths:
        print(
            "Found {} input videos: {} ...".format(
                len(input_video_paths),
                [os.path.basename(f) for f in input_video_paths[:5]],
            )
        )
        start = timer()
        for i, vid_path in enumerate(input_video_paths):
            output_path = os.path.join(
                output_path,
                os.path.basename(vid_path).replace(".", postfix + "."),
            )
            detect_video(yolo, vid_path, output_path=output_path)

        end = timer()
        print(
            "Processed {} videos in {:.1f}sec".format(
                len(input_video_paths), end - start
            )
        )
    # Close the current yolo session
    yolo.close_session()


        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        